# Copyright (c) 2021 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
import ctypes
from enum import IntEnum
import os
from typing import List, Optional, Sequence, cast

from numpy import floating
from numpy.typing import NDArray

from spinn_utilities.overrides import overrides

from pacman.model.placements import Placement
from pacman.model.graphs.common import Slice
from pacman.model.resources import AbstractSDRAM

from spinn_front_end_common.abstract_models import (
    AbstractGeneratesDataSpecification, AbstractRewritesDataSpecification)
from spinn_front_end_common.utilities.constants import BYTES_PER_WORD
from spinn_front_end_common.interface.ds import (
    DataSpecificationGenerator, DataSpecificationReloader)
from spinn_front_end_common.interface.provenance import ProvenanceWriter

from spynnaker.pyNN.utilities.utility_calls import get_n_bits
from spynnaker.pyNN.models.neuron.local_only import AbstractLocalOnly
from spynnaker.pyNN.models.neuron.neuron_data import NeuronData

from .population_machine_common import CommonRegions, PopulationMachineCommon
from .population_machine_neurons import (
    NeuronRegions, PopulationMachineNeurons, NeuronProvenance)
from .population_vertex import PopulationVertex


class LocalOnlyProvenance(ctypes.LittleEndianStructure):
    """
    Types of provenance and the DataType used to represent each.
    """
    _fields_ = [
        # The maximum number of spikes received in a time step
        ("max_spikes_per_timestep", ctypes.c_uint32),
        # The number of packets that were dropped due to being late
        ("n_spikes_dropped", ctypes.c_uint32),
        # A count of the times that the synaptic input circular buffers
        # overflowed
        ("n_spikes_lost_from_input", ctypes.c_uint32),
        # The maximum size of the spike input buffer during simulation
        ("max_size_input_buffer", ctypes.c_uint32)
    ]

    N_ITEMS = len(_fields_)


class MainProvenance(ctypes.LittleEndianStructure):
    """
    Provenance items from synapse processing.
    """
    _fields_ = [
        # the maximum number of background tasks queued
        ("max_background_queued", ctypes.c_uint32),
        # the number of times the background queue overloaded
        ("n_background_overloads", ctypes.c_uint32)
    ]

    N_ITEMS = len(_fields_)


class PopulationMachineLocalOnlyCombinedVertex(
        PopulationMachineCommon,
        PopulationMachineNeurons,
        AbstractGeneratesDataSpecification,
        AbstractRewritesDataSpecification):
    """
    A machine vertex for PyNN Populations.
    """

    __slots__ = (
        "__key",
        "__ring_buffer_shifts",
        "__weight_scales",
        "__slice_index",
        "__neuron_data",
        "__max_atoms_per_core",
        "__regenerate_data")

    # log_n_neurons, log_n_synapse_types, log_max_delay, input_buffer_size,
    # clear_input_buffer
    LOCAL_ONLY_SIZE = 5 * BYTES_PER_WORD

    INPUT_BUFFER_FULL_NAME = "Times_the_input_buffer_lost_packets"
    N_LATE_SPIKES_NAME = "Number_of_late_spikes"
    MAX_FILLED_SIZE_OF_INPUT_BUFFER_NAME = "Max_filled_size_input_buffer"
    MAX_SPIKES_PER_TIME_STEP_NAME = "Max_spikes_per_time_step"
    BACKGROUND_OVERLOADS_NAME = "Times_the_background_queue_overloaded"
    BACKGROUND_MAX_QUEUED_NAME = "Max_backgrounds_queued"

    class REGIONS(IntEnum):
        """
        Regions for populations.
        """
        SYSTEM = 0
        PROVENANCE_DATA = 1
        PROFILING = 2
        RECORDING = 3
        CORE_PARAMS = 4
        NEURON_PARAMS = 5
        CURRENT_SOURCE_PARAMS = 6
        NEURON_RECORDING = 7
        LOCAL_ONLY = 8
        LOCAL_ONLY_PARAMS = 9
        NEURON_BUILDER = 10
        INITIAL_VALUES = 11

    # Regions for this vertex used by common parts
    COMMON_REGIONS = CommonRegions(
        REGIONS.SYSTEM,
        REGIONS.PROVENANCE_DATA,
        REGIONS.PROFILING,
        REGIONS.RECORDING)

    # Regions for this vertex used by neuron parts
    NEURON_REGIONS = NeuronRegions(
        REGIONS.CORE_PARAMS,
        REGIONS.NEURON_PARAMS,
        REGIONS.CURRENT_SOURCE_PARAMS,
        REGIONS.NEURON_RECORDING,
        REGIONS.NEURON_BUILDER,
        REGIONS.INITIAL_VALUES)

    _PROFILE_TAG_LABELS = {
        0: "TIMER",
        1: "DMA_READ",
        2: "INCOMING_SPIKE"}

    def __init__(
            self, sdram: AbstractSDRAM, label: str,
            app_vertex: PopulationVertex, vertex_slice: Slice,
            slice_index: int, ring_buffer_shifts: Sequence[int],
            weight_scales: NDArray[floating], neuron_data: NeuronData,
            max_atoms_per_core: int):
        """
        :param sdram: The SDRAM used by the vertex
        :param label: The label of the vertex
        :param app_vertex: The associated application vertex
        :param vertex_slice: The slice of the population that this implements
        :param slice_index:
            The index of the slice in the ordered list of slices
        :param ring_buffer_shifts:
            The shifts to apply to convert ring buffer values to S1615 values
        :param weight_scales:
            The scaling to apply to weights to store them in the synapses
        :param neuron_data:
            The handler of neuron data
        :param max_atoms_per_core:
            The maximum number of atoms per core
        """
        super().__init__(
            label, app_vertex, vertex_slice, sdram,
            self.COMMON_REGIONS,
            NeuronProvenance.N_ITEMS +
            LocalOnlyProvenance.N_ITEMS + MainProvenance.N_ITEMS,
            self._PROFILE_TAG_LABELS, self.__get_binary_file_name(app_vertex))
        self.__key: Optional[int] = None
        self.__slice_index = slice_index
        self.__ring_buffer_shifts = ring_buffer_shifts
        self.__weight_scales = weight_scales
        self.__neuron_data = neuron_data
        self.__max_atoms_per_core = max_atoms_per_core
        self.__regenerate_data = False

    @property
    def _vertex_slice(self) -> Slice:
        return self.vertex_slice

    @property
    @overrides(PopulationMachineNeurons._slice_index)
    def _slice_index(self) -> int:
        return self.__slice_index

    @property
    @overrides(PopulationMachineNeurons._key)
    def _key(self) -> int:
        if self.__key is None:
            raise KeyError("'_key' not yet set")
        return self.__key

    @property
    @overrides(PopulationMachineNeurons._has_key)
    def _has_key(self) -> bool:
        return self.__key is not None

    @overrides(PopulationMachineNeurons._set_key)
    def _set_key(self, key: int) -> None:
        self.__key = key

    @property
    @overrides(PopulationMachineNeurons._neuron_regions)
    def _neuron_regions(self) -> NeuronRegions:
        return self.NEURON_REGIONS

    @property
    @overrides(PopulationMachineNeurons._neuron_data)
    def _neuron_data(self) -> NeuronData:
        return self.__neuron_data

    @property
    @overrides(PopulationMachineNeurons._max_atoms_per_core)
    def _max_atoms_per_core(self) -> int:
        return self.__max_atoms_per_core

    @property
    def __synapse_dynamics(self) -> AbstractLocalOnly:
        # Precondition for construction of this class instance
        return cast(AbstractLocalOnly, self._pop_vertex.synapse_dynamics)

    @staticmethod
    def __get_binary_file_name(app_vertex: PopulationVertex) -> str:
        """
        Get the local binary filename for this vertex.  Static because at
        the time this is needed, the local app_vertex is not set.

        :param app_vertex:
            The associated application vertex
        """
        # Split binary name into title and extension
        name, ext = os.path.splitext(app_vertex.neuron_impl.binary_name)

        # Reunite title and extension and return
        return name + app_vertex.synapse_executable_suffix + ext

    @overrides(PopulationMachineCommon.parse_extra_provenance_items)
    def parse_extra_provenance_items(
            self, label: str, x: int, y: int, p: int,
            provenance_data: Sequence[int]) -> None:
        proc_offset = NeuronProvenance.N_ITEMS
        end_proc_offset = proc_offset + LocalOnlyProvenance.N_ITEMS
        self._parse_neuron_provenance(
            x, y, p, provenance_data[:NeuronProvenance.N_ITEMS])
        self._parse_local_only_provenance(
            label, x, y, p, provenance_data[proc_offset:end_proc_offset])

        main_prov = MainProvenance(*provenance_data[-MainProvenance.N_ITEMS:])
        with ProvenanceWriter() as db:
            db.insert_core(
                x, y, p, self.BACKGROUND_MAX_QUEUED_NAME,
                main_prov.max_background_queued)
            if main_prov.max_background_queued > 1:
                db.insert_report(
                    f"A maximum of {main_prov.max_background_queued}"
                    f" background tasks were queued on {label}.  Try"
                    " increasing the time_scale_factor located within the"
                    " .spynnaker.cfg file or in the pynn.setup() method.")
            db.insert_core(
                x, y, p, self.BACKGROUND_OVERLOADS_NAME,
                main_prov.n_background_overloads)

            if main_prov.n_background_overloads > 0:
                db.insert_report(
                    "The background queue overloaded "
                    f"{main_prov.n_background_overloads} times on {label}."
                    " Try increasing the time_scale_factor located within"
                    " the .spynnaker.cfg file or in the pynn.setup() method.")

    @overrides(PopulationMachineCommon.get_recorded_region_ids)
    def get_recorded_region_ids(self) -> List[int]:
        ids = self._pop_vertex.neuron_recorder.recorded_ids_by_slice(
            self.vertex_slice)
        ids.extend(self._pop_vertex.synapse_recorder.recorded_ids_by_slice(
            self.vertex_slice))
        return ids

    @overrides(AbstractGeneratesDataSpecification.generate_data_specification)
    def generate_data_specification(self, spec: DataSpecificationGenerator,
                                    placement: Placement) -> None:
        rec_regions = self._pop_vertex.neuron_recorder.get_region_sizes(
            self.vertex_slice)
        rec_regions.extend(self._pop_vertex.synapse_recorder.get_region_sizes(
            self.vertex_slice))
        self._write_common_data_spec(spec, rec_regions)

        self._write_neuron_data_spec(spec, self.__ring_buffer_shifts)

        self.__write_local_only_data(spec)

        self.__synapse_dynamics.write_parameters(
            spec, self.REGIONS.LOCAL_ONLY_PARAMS, self,
            self.__weight_scales)

        # End the writing of this specification:
        spec.end_specification()

    def __write_local_only_data(
            self, spec: DataSpecificationGenerator) -> None:
        spec.reserve_memory_region(
            self.REGIONS.LOCAL_ONLY, self.LOCAL_ONLY_SIZE, "local_only")
        spec.switch_write_focus(self.REGIONS.LOCAL_ONLY)
        log_n_max_atoms = get_n_bits(self._max_atoms_per_core)
        log_n_synapse_types = get_n_bits(
            self._pop_vertex.neuron_impl.get_n_synapse_types())
        # Find the maximum delay
        max_delay = self._pop_vertex.splitter.max_support_delay()

        spec.write_value(log_n_max_atoms)
        spec.write_value(log_n_synapse_types)
        spec.write_value(get_n_bits(max_delay))
        spec.write_value(self._pop_vertex.incoming_spike_buffer_size)
        spec.write_value(int(self._pop_vertex.drop_late_spikes))

    @overrides(AbstractRewritesDataSpecification.regenerate_data_specification)
    def regenerate_data_specification(self, spec: DataSpecificationReloader,
                                      placement: Placement) -> None:
        self._rewrite_neuron_data_spec(spec)

        # close spec
        spec.end_specification()

    @overrides(AbstractRewritesDataSpecification.reload_required)
    def reload_required(self) -> bool:
        return self.__regenerate_data

    @overrides(AbstractRewritesDataSpecification.set_reload_required)
    def set_reload_required(self, new_value: bool) -> None:
        self.__regenerate_data = new_value

    def _parse_local_only_provenance(
            self, label: str, x: int, y: int, p: int,
            provenance_data: Sequence[int]) -> None:
        """
        Extract and yield local-only provenance.

        :param label: The label of the node
        :param x: x coordinate of the chip where this core
        :param y: y coordinate of the core where this core
        :param p: virtual id of the core
        :param provenance_data: A list of data items to interpret
        """
        prov = LocalOnlyProvenance(*provenance_data)

        with ProvenanceWriter() as db:
            db.insert_core(
                x, y, p, self.MAX_SPIKES_PER_TIME_STEP_NAME,
                prov.max_spikes_per_timestep)
            db.insert_core(
                x, y, p, self.INPUT_BUFFER_FULL_NAME,
                prov.n_spikes_lost_from_input)
            db.insert_core(
                x, y, p, self.N_LATE_SPIKES_NAME,
                prov.n_spikes_dropped)
            db.insert_core(
                x, y, p, self.MAX_FILLED_SIZE_OF_INPUT_BUFFER_NAME,
                prov.max_size_input_buffer)

            if prov.n_spikes_lost_from_input > 0:
                db.insert_report(
                    f"The input buffer for {label} lost packets on "
                    f"{prov.n_spikes_lost_from_input} occasions. This is "
                    "often sign that the system is running too quickly for "
                    "the number of neurons per core.  Please increase the "
                    "timer_tic or time_scale_factor or decrease the number "
                    "of neurons per core.")

            if prov.n_spikes_dropped > 0:
                if self._pop_vertex.drop_late_spikes:
                    db.insert_report(
                        f"On {label}, {prov.n_spikes_dropped} packets were "
                        "dropped from the input buffer, because they arrived "
                        "too late to be processed in a given time step. Try "
                        "increasing the time_scale_factor located within the "
                        ".spynnaker.cfg file or in the pynn.setup() method.")
                else:
                    db.insert_report(
                        f"On {label}, {prov.n_spikes_dropped} packets arrived "
                        "too late to be processed in a given time step. Try "
                        "increasing the time_scale_factor located within the "
                        ".spynnaker.cfg file or in the pynn.setup() method.")

    @overrides(PopulationMachineNeurons.set_do_neuron_regeneration)
    def set_do_neuron_regeneration(self) -> None:
        self.__regenerate_data = True



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional
from pyNN.random import RandomDistribution
from spinn_utilities.abstract_base import AbstractBase, abstractmethod


class AbstractRandomStats(object, metaclass=AbstractBase):
    """
    Statistics about PyNN `~spynnaker.pyNN.RandomDistribution` objects.
    """
    __slots__ = ()

    @abstractmethod
    def cdf(self, dist: RandomDistribution, v: float) -> float:
        """
        :returns: The cumulative distribution function value for the value `v`.
        """
        raise NotImplementedError

    @abstractmethod
    def ppf(self, dist: RandomDistribution, p: float) -> float:
        """
        :returns: The percent point function value for the probability `p`.
        """
        raise NotImplementedError

    @abstractmethod
    def mean(self, dist: RandomDistribution) -> float:
        """
        :returns: The mean of the distribution.
        """
        raise NotImplementedError

    @abstractmethod
    def std(self, dist: RandomDistribution) -> float:
        """
        :returns: The standard deviation of the distribution.
        """
        raise NotImplementedError

    @abstractmethod
    def var(self, dist: RandomDistribution) -> float:
        """
        :returns: The variance of the distribution.
        """
        raise NotImplementedError

    @abstractmethod
    def high(self, distribution: RandomDistribution) -> Optional[float]:
        """
        :returns: The high cut-off value of the distribution, or `None` if the
        distribution is unbounded.
        """
        raise NotImplementedError

    @abstractmethod
    def low(self, distribution: RandomDistribution) -> Optional[float]:
        """
        :returns: The low cut-off value of the distribution, or `None` if the
        distribution is unbounded.
        """
        raise NotImplementedError



# Copyright (c) 2020 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .abstract_spynnaker_splitter_delay import AbstractSpynnakerSplitterDelay
from .spynnaker_splitter_selector import spynnaker_splitter_selector
from .splitter_delay_vertex_slice import SplitterDelayVertexSlice
from .splitter_poisson_delegate import SplitterPoissonDelegate
from .splitter_population_vertex import SplitterPopulationVertex
from .splitter_population_vertex_fixed import SplitterPopulationVertexFixed
from .splitter_population_vertex_neurons_synapses import (
    SplitterPopulationVertexNeuronsSynapses)
from .abstract_supports_one_to_one_sdram_input import (
    AbstractSupportsOneToOneSDRAMInput)

__all__ = [
    'AbstractSpynnakerSplitterDelay', 'AbstractSupportsOneToOneSDRAMInput',
    'SplitterDelayVertexSlice', 'SplitterPoissonDelegate',
    'SplitterPopulationVertex', 'SplitterPopulationVertexFixed',
    'SplitterPopulationVertexNeuronsSynapses',
    'spynnaker_splitter_selector']



# Copyright (c) 2023 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Final
from numpy import dtype, float64, uint32, void
from numpy.typing import NDArray
from typing_extensions import TypeAlias

#: Type model of the basic configuration data of a connector
NUMPY_CONNECTORS_DTYPE: Final = dtype(
    [("source", uint32), ("target", uint32),
     ("weight", float64), ("delay", float64)])

#: Type of connections. The dtype is actually
#: :py:const:`NUMPY_CONNECTORS_DTYPE` but we cannot currently express that in
#: the overall array type due to Numpy typing limitations.
ConnectionsArray: TypeAlias = NDArray[void]



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from time import sleep
from typing import Optional

from spinn_utilities.overrides import overrides
from spinn_utilities.log import FormatAdapter

from spinn_front_end_common.utility_models import MultiCastCommand

from spynnaker.pyNN.external_devices_models import AbstractEthernetTranslator
from spynnaker.pyNN.protocols import (
    MunichIoEthernetProtocol, munich_io_spinnaker_link_protocol)
from spynnaker.pyNN.protocols.munich_io_spinnaker_link_protocol import (
    MunichIoSpiNNakerLinkProtocol)

from .push_bot_wifi_connection import PushBotWIFIConnection

logger = FormatAdapter(logging.getLogger(__name__))


def _signed_int(uint_value: Optional[int]) -> int:
    if uint_value is None:
        return 0
    if uint_value > (2 ** 31):
        return uint_value - (2 ** 32)
    return uint_value


class PushBotTranslator(AbstractEthernetTranslator):
    """
    Translates packets between PushBot Multicast packets and PushBot
    Wi-Fi Commands.
    """
    __slots__ = (
        "__protocol",
        "__pushbot_wifi_connection")

    def __init__(self, protocol: MunichIoSpiNNakerLinkProtocol,
                 pushbot_wifi_connection: PushBotWIFIConnection):
        """
        :param protocol:
            The instance of the PushBot protocol to get keys from
        :param pushbot_wifi_connection:
            A Wi-Fi connection to the PushBot
        """
        self.__protocol = protocol
        self.__pushbot_wifi_connection = pushbot_wifi_connection

    @overrides(AbstractEthernetTranslator.translate_control_packet)
    def translate_control_packet(
            self, multicast_packet: MultiCastCommand) -> None:
        # pylint: disable=too-many-statements, too-many-branches
        key = multicast_packet.key
        # disable retina
        if key == self.__protocol.disable_retina_key:
            logger.debug("Sending retina disable")
            self.__pushbot_wifi_connection.send(
                MunichIoEthernetProtocol.disable_retina())
            sleep(0.1)

        # set retina key (which doesn't do much for Ethernet)
        elif key == self.__protocol.set_retina_transmission_key:
            logger.debug("Sending retina enable")
            self.__pushbot_wifi_connection.send(
                MunichIoEthernetProtocol.set_retina_transmission(
                    munich_io_spinnaker_link_protocol.GET_RETINA_PAYLOAD_VALUE(
                        multicast_packet.payload or 0)))
            self.__pushbot_wifi_connection.send(
                MunichIoEthernetProtocol.enable_retina())

        # motor 0 leaky velocity command
        elif key == self.__protocol.push_bot_motor_0_leaking_towards_zero_key:
            speed = _signed_int(multicast_packet.payload)
            logger.debug("Sending Motor 0 Leaky Velocity = {}", speed)
            self.__pushbot_wifi_connection.send(
                MunichIoEthernetProtocol.motor_0_leaky_velocity(speed))

        # motor 0 permanent velocity command
        elif key == self.__protocol.push_bot_motor_0_permanent_key:
            speed = _signed_int(multicast_packet.payload)
            logger.debug("Sending Motor 0 Velocity = {}", speed)
            self.__pushbot_wifi_connection.send(
                MunichIoEthernetProtocol.motor_0_permanent_velocity(speed))

        # motor 1 leaky velocity command
        elif key == self.__protocol.push_bot_motor_1_leaking_towards_zero_key:
            speed = _signed_int(multicast_packet.payload)
            logger.debug("Sending Motor 1 Leaky Velocity = {}", speed)
            self.__pushbot_wifi_connection.send(
                MunichIoEthernetProtocol.motor_1_leaky_velocity(speed))

        # motor 1 permanent velocity command
        elif key == self.__protocol.push_bot_motor_1_permanent_key:
            speed = _signed_int(multicast_packet.payload)
            logger.debug("Sending Motor 1 Velocity = {}", speed)
            self.__pushbot_wifi_connection.send(
                MunichIoEthernetProtocol.motor_1_permanent_velocity(speed))

        # laser total period command
        elif key == self.__protocol.push_bot_laser_config_total_period_key:
            period = _signed_int(multicast_packet.payload)
            logger.debug("Sending Laser Period = {}", period)
            self.__pushbot_wifi_connection.send(
                MunichIoEthernetProtocol.laser_total_period(period))

        # laser active time
        elif key == self.__protocol.push_bot_laser_config_active_time_key:
            time = _signed_int(multicast_packet.payload)
            logger.debug("Sending Laser Active Time = {}", time)
            self.__pushbot_wifi_connection.send(
                MunichIoEthernetProtocol.laser_active_time(time))

        # laser frequency
        elif key == self.__protocol.push_bot_laser_set_frequency_key:
            frequency = _signed_int(multicast_packet.payload)
            logger.debug("Sending Laser Frequency = {}", frequency)
            self.__pushbot_wifi_connection.send(
                MunichIoEthernetProtocol.laser_frequency(frequency))

        # led total period command
        elif key == self.__protocol.push_bot_led_total_period_key:
            period = _signed_int(multicast_packet.payload)
            logger.debug("Sending LED Period = {}", period)
            self.__pushbot_wifi_connection.send(
                MunichIoEthernetProtocol.led_total_period(period))

        # front led active time
        elif key == self.__protocol.push_bot_led_front_active_time_key:
            time = _signed_int(multicast_packet.payload)
            logger.debug("Sending Front LED Active Time = {}", time)
            self.__pushbot_wifi_connection.send(
                MunichIoEthernetProtocol.led_front_active_time(time))

        # back led active time
        elif key == self.__protocol.push_bot_led_back_active_time_key:
            time = _signed_int(multicast_packet.payload)
            logger.debug("Sending Back LED Active Time = {}", time)
            self.__pushbot_wifi_connection.send(
                MunichIoEthernetProtocol.led_back_active_time(time))

        # led frequency
        elif key == self.__protocol.push_bot_led_set_frequency_key:
            frequency = _signed_int(multicast_packet.payload)
            logger.debug("Sending LED Frequency = {}", frequency)
            self.__pushbot_wifi_connection.send(
                MunichIoEthernetProtocol.led_frequency(frequency))

        # speaker total period
        elif key == self.__protocol.push_bot_speaker_config_total_period_key:
            period = _signed_int(multicast_packet.payload)
            logger.debug("Sending Speaker Period = {}", period)
            self.__pushbot_wifi_connection.send(
                MunichIoEthernetProtocol.speaker_total_period(period))

        # speaker active time
        elif key == self.__protocol.push_bot_speaker_config_active_time_key:
            time = _signed_int(multicast_packet.payload)
            logger.debug("Sending Speaker Active Time = {}", time)
            self.__pushbot_wifi_connection.send(
                MunichIoEthernetProtocol.speaker_active_time(time))

        # speaker frequency
        elif key == self.__protocol.push_bot_speaker_set_tone_key:
            frequency = _signed_int(multicast_packet.payload)
            logger.debug("Sending Speaker Frequency = {}", frequency)
            self.__pushbot_wifi_connection.send(
                MunichIoEthernetProtocol.speaker_frequency(frequency))

        # motor enable
        elif (key == self.__protocol.enable_disable_motor_key and
              multicast_packet.payload == 1):
            logger.debug("Sending Motor Enable")
            self.__pushbot_wifi_connection.send(
                MunichIoEthernetProtocol.enable_motor())

        # motor disable
        elif (key == self.__protocol.enable_disable_motor_key and
              multicast_packet.payload == 0):
            logger.debug("Sending Motor Disable")
            self.__pushbot_wifi_connection.send(
                MunichIoEthernetProtocol.disable_motor())

        # detecting set mode (which has no context in Ethernet protocol
        elif key == self.__protocol.set_mode().key:
            logger.debug("Ignoring set mode command")

        # otherwise no idea what command is, so raise warning and ignore
        else:
            logger.warning("Unknown PushBot command: {}", multicast_packet)



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from typing import Iterable, Optional, TYPE_CHECKING

from numpy import floating
from numpy.typing import NDArray
from typing_extensions import TypeAlias

from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from pacman.model.graphs.common import Slice
from spinn_front_end_common.interface.ds import DataSpecificationBase

if TYPE_CHECKING:
    from spynnaker.pyNN.models.neuron.synapse_dynamics.types import (
        ConnectionsArray)
    from spynnaker.pyNN.models.neuron.structural_plasticity.synaptogenesis\
        .partner_selection import AbstractPartnerSelection
    from spynnaker.pyNN.models.neuron.structural_plasticity.synaptogenesis\
        .formation import AbstractFormation
    from spynnaker.pyNN.models.neuron.structural_plasticity.synaptogenesis\
        .elimination import AbstractElimination
    from spynnaker.pyNN.models.projection import Projection
    from spynnaker.pyNN.models.neuron.synaptic_matrices import SynapticMatrices
    from spynnaker.pyNN.models.neuron import PopulationVertex
    from spynnaker.pyNN.models.neural_projections import (
        ProjectionApplicationEdge, SynapseInformation)

# see https://github.com/SpiNNakerManchester/sPyNNaker/issues/1427
#: :meta private:
InitialDelay: TypeAlias = float


class AbstractSynapseDynamicsStructural(object, metaclass=AbstractBase):
    """
    Base class for synapse dynamics that structural plasticity understands.
    """
    __slots__ = ()

    @abstractmethod
    def get_structural_parameters_sdram_usage_in_bytes(
            self, incoming_projections: Iterable[Projection],
            n_neurons: int) -> int:
        """
        Get the size of the structural parameters.

        .. note::
            At the Application level this will be an estimate.

        :param incoming_projections:
            The projections that target the vertex in question
        :param n_neurons:
        :return: the size of the parameters, in bytes
        :raises PacmanInvalidParameterException:
            If the parameters make no sense.
        """
        raise NotImplementedError

    @abstractmethod
    def write_structural_parameters(
            self, spec: DataSpecificationBase, region: int,
            weight_scales: NDArray[floating],
            app_vertex: PopulationVertex,
            vertex_slice: Slice, synaptic_matrices: SynapticMatrices) -> None:
        """
        Write structural plasticity parameters.

        :param spec: The data specification to write to
        :param region: region ID
        :param weight_scales: Weight scaling for each synapse type
        :param app_vertex: The target application vertex
        :param vertex_slice: The slice of the target vertex to generate for
        :param synaptic_matrices: The synaptic matrices for this vertex
        """
        raise NotImplementedError

    @abstractmethod
    def set_connections(
            self, connections: ConnectionsArray, post_vertex_slice: Slice,
            app_edge: ProjectionApplicationEdge,
            synapse_info: SynapseInformation) -> None:
        """
        Set connections for structural plasticity.

        :param connections:
        :param post_vertex_slice:
        :param app_edge:
        :param synapse_info:
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def f_rew(self) -> float:
        """
        The frequency of rewiring.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def s_max(self) -> int:
        """
        The maximum number of synapses.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def with_replacement(self) -> bool:
        """
        Whether to allow replacement when creating synapses.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def seed(self) -> Optional[int]:
        """
        The seed to control the randomness.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def initial_weight(self) -> float:
        """
        The weight of a formed connection.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def initial_delay(self) -> InitialDelay:
        """
        The delay of a formed connection.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def partner_selection(self) -> AbstractPartnerSelection:
        """
        The partner selection rule.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def formation(self) -> AbstractFormation:
        """
        The formation rule.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def elimination(self) -> AbstractElimination:
        """
        The elimination rule.
        """
        raise NotImplementedError

    @abstractmethod
    def check_initial_delay(self, max_delay_ms: int) -> None:
        """
        Check that delays can be done without delay extensions.

        :param max_delay_ms: The maximum delay supported, in milliseconds
        :raises Exception: if the delay is out of range
        """
        raise NotImplementedError

    @abstractmethod
    def get_max_rewires_per_ts(self) -> int:
        """
        :returns: The max number of rewires per timestep.
        """
        raise NotImplementedError



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spynnaker.pyNN.exceptions import SpynnakerException
from spynnaker.pyNN.models.defaults import (
    AbstractProvidesDefaults, default_initial_values)
from spynnaker.pyNN.models.neuron.implementations import ModelParameter


class IFFacetsConductancePopulation(AbstractProvidesDefaults):
    """
    Leaky integrate and fire neuron with conductance-based synapses and
    fixed threshold as it is resembled by the FACETS Hardware Stage 1.

    .. warning::
        Not currently supported by the tool chain.
    """

    # noinspection PyPep8Naming
    @default_initial_values({"v"})
    def __init__(
            self, g_leak: ModelParameter = 40.0,
            tau_syn_E: ModelParameter = 30.0,
            tau_syn_I: ModelParameter = 30.0, v_thresh: ModelParameter = -55.0,
            v_rest: ModelParameter = -65.0, e_rev_I: ModelParameter = -80,
            v_reset: ModelParameter = -80.0, v: ModelParameter = -65.0):
        """
        :param g_leak:
        :param tau_syn_E:
        :param tau_syn_I:
        :param v_thresh:
        :param v_rest:
        :param e_rev_I:
        :param v_reset:
        :param v:
        """
        # pylint: disable=unused-argument, invalid-name
        raise SpynnakerException(
            "This neuron model is currently not supported by the tool chain")



# Copyright (c) 2022 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import (auto, Enum)


class BufferDataType(Enum):
    """
    Different functions to retrieve the data.

    This class is designed to used internally by NeoBufferDatabase
    """
    NEURON_SPIKES = (auto())
    EIEIO_SPIKES = (auto())
    MULTI_SPIKES = (auto())
    MATRIX = (auto())
    REWIRES = (auto())
    NOT_NEO = (auto())

    def __str__(self) -> str:
        return self.name



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
import functools
import logging
from typing import (
    Any, Callable, Dict, List, Optional, Sequence, Tuple, Union,
    cast, TYPE_CHECKING)

import numpy
from numpy import void
from numpy.typing import NDArray
from typing_extensions import Literal, Never, TypeAlias

from pyNN.recording.files import BaseFile
from pyNN.space import Space as PyNNSpace

from spinn_utilities.config_holder import get_config_bool
from spinn_utilities.log import FormatAdapter
from spinn_utilities.logger_utils import warn_once

from pacman.model.graphs.application import ApplicationVertex

from spinn_front_end_common.utilities.exceptions import ConfigurationException

from spynnaker.pyNN.data import SpynnakerDataView
from spynnaker.pyNN.models.abstract_models import (
    AbstractAcceptsIncomingSynapses)
from spynnaker.pyNN.models.neural_projections import (
    SynapseInformation, ProjectionApplicationEdge)
from spynnaker.pyNN.models.neural_projections.connectors import (
    FromListConnector)
from spynnaker.pyNN.models.neuron import (
    PopulationVertex, ConnectionHolder)
from spynnaker.pyNN.models.populations import Population, PopulationView
from spynnaker.pyNN.models.neuron.synapse_dynamics import (
    SynapseDynamicsStatic, AbstractHasParameterNames)
from spynnaker.pyNN.models.spike_source import SpikeSourcePoissonVertex
from spynnaker._version import __version__

if TYPE_CHECKING:
    from spynnaker.pyNN.models.neural_projections.connectors import (
        AbstractConnector)
    from spynnaker.pyNN.models.neuron.synapse_dynamics import (
        AbstractSynapseDynamics)
    _Pop: TypeAlias = Union[Population, PopulationView]

logger = FormatAdapter(logging.getLogger(__name__))


class Projection(object):
    """
    A container for all the connections of a given type (same synapse type and
    plasticity mechanisms) between two populations, together with methods to
    set parameters of those connections, including of plasticity mechanisms.
    """
    # "format" param name defined by PyNN/
    # pylint: disable=redefined-builtin
    __slots__ = (
        "__projection_edge",
        "__synapse_information",
        "__virtual_connection_list",
        "__label"
    )

    def __init__(
            self, pre_synaptic_population: _Pop,
            post_synaptic_population: _Pop, connector: AbstractConnector,
            synapse_type: Optional[AbstractSynapseDynamics] = None,
            source: None = None, receptor_type: str = "excitatory",
            space: Optional[PyNNSpace] = None,
            label: Optional[str] = None,
            download_synapses: bool = False,
            partition_id: Optional[str] = None):
        """
        :param pre_synaptic_population:
        :param post_synaptic_population:
        :param connector:
        :param synapse_type:
        :param source: Unsupported; must be `None`
        :param receptor_type:
        :param space:
        :param label:
        :param download_synapses:
        """
        if source is not None:
            raise NotImplementedError(
                f"sPyNNaker {__version__} does not yet support "
                "multi-compartmental cells.")

        pre_is_view = self.__check_population(pre_synaptic_population)
        post_is_view = self.__check_population(post_synaptic_population)

        # set default label
        if label is None:
            # set the projection's label to a default (maybe non-unique!)
            self.__label = (
                f"from pre {pre_synaptic_population.label} "
                f"to post {post_synaptic_population.label} "
                f"with connector {connector}")
            # give an auto generated label for the underlying edge
            label = f"projection edge " \
                    f"{SpynnakerDataView.get_next_none_labelled_edge_number()}"
        else:
            self.__label = label

        # Handle default synapse type
        if synapse_type is None:
            synapse_dynamics: AbstractSynapseDynamics = SynapseDynamicsStatic()
        else:
            synapse_dynamics = synapse_type

        # set the space function as required
        if space is None:
            space = PyNNSpace()
        connector.set_space(space)

        pre_vertex = pre_synaptic_population._vertex
        post_vertex = cast(PopulationVertex, post_synaptic_population._vertex)

        if not isinstance(post_vertex, AbstractAcceptsIncomingSynapses):
            raise ConfigurationException(
                "postsynaptic population is not designed to receive"
                " synaptic projections")

        # sort out synapse type
        synapse_id = post_vertex.get_synapse_id_by_target(receptor_type)
        synapse_id_from_dynamics = False
        if synapse_id is None:
            synapse_id = synapse_dynamics.get_synapse_id_by_target(
                receptor_type)
            synapse_id_from_dynamics = True
        if synapse_id is None:
            raise ConfigurationException(
                f"Synapse target {receptor_type} not found "
                f"in {post_synaptic_population.label}")

        # as a from-list connector can have plastic parameters, grab those (
        # if any) and add them to the synapse dynamics object
        if isinstance(connector, FromListConnector):
            connector._apply_parameters_to_synapse_type(synapse_dynamics)

        # set the plasticity dynamics for the post pop (allows plastic stuff
        #  when needed)
        post_vertex.set_synapse_dynamics(synapse_dynamics)

        # Set and store synapse information for future processing
        self.__synapse_information = SynapseInformation(
            connector, pre_synaptic_population, post_synaptic_population,
            pre_is_view, post_is_view, synapse_dynamics,
            synapse_id, receptor_type, synapse_id_from_dynamics,
            synapse_dynamics.weight, synapse_dynamics.delay,
            download_synapses, partition_id)

        # Set projection information in connector
        connector.set_projection_information(self.__synapse_information)

        # Find out if there is an existing edge between the populations
        edge_to_merge = self._find_existing_edge(
            pre_vertex, post_vertex, self.__synapse_information.partition_id)
        if edge_to_merge is not None:
            # If there is an existing edge, add the connector
            edge_to_merge.add_synapse_information(self.__synapse_information)
            self.__projection_edge = edge_to_merge
        else:
            # If there isn't an existing edge, create a new one and add it
            self.__projection_edge = ProjectionApplicationEdge(
                pre_vertex, post_vertex, self.__synapse_information,
                label=label)
            SpynnakerDataView.add_edge(
                self.__projection_edge,
                self.__synapse_information.partition_id)

        # Ensure the connector is happy
        synapse_dynamics.validate_connection(
            self.__projection_edge, self.__synapse_information)

        # add projection to the SpiNNaker control system
        SpynnakerDataView.add_projection(self)

        # If there is a virtual board, we need to hold the data in case the
        # user asks for it
        self.__virtual_connection_list: Optional[List[NDArray[void]]] = None
        if get_config_bool("Machine", "virtual_board"):
            self.__virtual_connection_list = list()
            self.__synapse_information.add_pre_run_connection_holder(
                ConnectionHolder(
                    None, False, pre_vertex.n_atoms, post_vertex.n_atoms,
                    self.__virtual_connection_list))

        # If the target is a population, add to the list of incoming
        # projections
        if isinstance(post_vertex, PopulationVertex):
            post_vertex.add_incoming_projection(self)

        # If the source is a poisson, add to the list of outgoing projections
        if isinstance(pre_vertex, SpikeSourcePoissonVertex):
            pre_vertex.add_outgoing_projection(self)

    @staticmethod
    def __check_population(param: _Pop) -> bool:
        """
        :param  param:
        :return: Whether the parameter is a view
        """
        if isinstance(param, Population):
            # Projections definitely work from Populations
            return False
        if not isinstance(param, PopulationView):
            raise ConfigurationException(
                f"Unexpected parameter type {type(param)}. "
                "Expected Population")
        # Check whether the array is contiguous or not
        if not param._is_contiguous:  # pylint: disable=protected-access
            raise NotImplementedError(
                "Projections over views only work on contiguous arrays, "
                "e.g. view = pop[n:m], not view = pop[n,m]")
        # Projection is compatible with PopulationView
        return True

    def get(self, attribute_names: Union[str, Sequence[str]],
            format: str,  # @ReservedAssignment
            gather: Literal[True] = True, with_address: bool = True,
            multiple_synapses: Literal['last'] = 'last') -> ConnectionHolder:
        """
        Get a parameter/attribute of the projection.

        .. note::
            SpiNNaker always gathers.

        :param attribute_names: list of attributes to gather
        :param format: ``"list"`` or ``"array"``
        :param gather: Ignored. Purely for PyNN compatibility
        :param with_address:
            True if the source and target are to be included
        :param multiple_synapses:
            What to do with the data if format="array" and if the multiple
            source-target pairs with the same values exist.  Currently only
            "last" is supported
        :return: values selected
        """
        if not gather:
            logger.warning("sPyNNaker always gathers from every core.")
        if multiple_synapses != 'last':
            raise ConfigurationException(
                "sPyNNaker only recognises multiple_synapses == last")
        an = [attribute_names] if isinstance(attribute_names, str) else list(
            attribute_names)

        return self.__get_data(an, format, with_address, notify=None)

    def save(
            self, attribute_names: Union[str, Sequence[str]],
            file: Union[str, BaseFile],
            format: str = 'list',  # @ReservedAssignment
            gather: Literal[True] = True, with_address: bool = True) -> None:
        """
        Print synaptic attributes (weights, delays, etc.) to file. In the
        array format, zeros are printed for non-existent connections.
        Values will be expressed in the standard PyNN units (i.e.,
        millivolts, nanoamps, milliseconds, microsiemens, nanofarads,
        event per second).

        .. note::
            SpiNNaker always gathers.

        :param attribute_names:
        :param file: filename or open handle (which will be closed)
        :param format:
        :param gather: Ignored. Purely for PyNN compatibility
        :param with_address:
        """
        if not gather:
            warn_once(
                logger, "sPyNNaker only supports gather=True. We will run "
                "as if gather was set to True.")
        if isinstance(attribute_names, str):
            attribute_names = [attribute_names]
        else:
            attribute_names = list(attribute_names)
        if len(attribute_names) == 1 and attribute_names[0] in {
                'all', 'connections'}:
            sd = self._projection_edge.post_vertex.synapse_dynamics
            if isinstance(sd, AbstractHasParameterNames):
                attribute_names = list(sd.get_parameter_names())
            else:
                attribute_names = []
        metadata = {"columns": attribute_names}
        if with_address:
            metadata["columns"] = ["i", "j"] + list(metadata["columns"])
        self.__get_data(
            attribute_names, format, with_address,
            notify=functools.partial(self.__save_callback, file, metadata))

    def __get_data(
            self, attribute_names: List[str],
            format: str,  # @ReservedAssignment
            with_address: bool,
            notify: Optional[Callable[[ConnectionHolder], None]]
            ) -> ConnectionHolder:
        """
        Internal data getter to add notify option.

        :param attribute_names: list of attributes to gather
        :param format: ``"list"`` or ``"array"``
        :param with_address:
        :param notify:
        :return: values selected
        """
        # fix issue with 1 versus many
        if isinstance(attribute_names, str):
            attribute_names = [attribute_names]

        data_items: List[str] = list()
        if format != "list":
            with_address = False
        if with_address:
            data_items.append("source")
            data_items.append("target")
            if "source" in attribute_names:
                logger.warning(
                    "Ignoring request to get source as with_address=True. ")
                attribute_names.remove("source")
            if "target" in attribute_names:
                logger.warning(
                    "Ignoring request to get target as with_address=True. ")
                attribute_names.remove("target")

        # Split out attributes in to standard versus synapse dynamics data
        fixed_values: List[Tuple[str, int]] = list()
        for attribute in attribute_names:
            data_items.append(attribute)
            synapse_dynamics = self._synapse_information.synapse_dynamics
            if attribute not in synapse_dynamics.get_synapse_parameter_names():
                value = synapse_dynamics.get_value(attribute)
                fixed_values.append((attribute, value))

        # Return the connection data
        return self._get_synaptic_data(
            format == "list", data_items, fixed_values, notify=notify)

    @staticmethod
    def __save_callback(
            save_file: Union[str, BaseFile],
            metadata: Dict[str, Any], data: ConnectionHolder) -> None:
        """
        :param save_file:
        :param metadata:
        :param data:
        """
        # Convert structured array to normal numpy array
        if hasattr(data, "dtype") and hasattr(data.dtype, "names"):
            dtype = [(name, "<f8") for name in data.dtype.names]
            data = data.astype(dtype)
        npdata = numpy.nan_to_num(cast(NDArray, data))
        if isinstance(save_file, str):
            data_file = open(save_file, mode='wb')
        else:
            data_file = save_file
        try:
            header_lines = [
                f"# {key} = {value}" for key, value in metadata.items()]
            header = "\n".join(header_lines) + '\n'
            data_file.write(header.encode('utf-8'))
            # write data
            numpy.savetxt(data_file, npdata, delimiter='\t')
            data_file.close()
        finally:
            data_file.close()

    @property
    def pre(self) -> _Pop:
        """
        The pre-population or population view.
        """
        return self._synapse_information.pre_population

    @property
    def post(self) -> _Pop:
        """
        The post-population or population view.
        """
        return self._synapse_information.post_population

    @property
    def label(self) -> str:
        """
        Label of this Projection
        """
        return self.__label

    def __repr__(self) -> str:
        return f"projection {self.__label}"

    # -----------------------------------------------------------------

    @property
    def _synapse_information(self) -> SynapseInformation:
        return self.__synapse_information

    @property
    def _projection_edge(self) -> ProjectionApplicationEdge:
        return self.__projection_edge

    def _find_existing_edge(
            self, pre_synaptic_vertex: ApplicationVertex,
            post_synaptic_vertex: ApplicationVertex,
            partition_id: str) -> Optional[ProjectionApplicationEdge]:
        """
        Searches though the graph's edges to locate any
        edge which has the same post- and pre- vertex

        :param pre_synaptic_vertex: the source vertex of the multapse
        :param post_synaptic_vertex: The destination vertex of the multapse
        :param partition_id: The partition ID of the edge to find
        :return: `None` or the edge going to these vertices.
        """
        # Find edges ending at the postsynaptic vertex
        partitions = (
            SpynnakerDataView.get_outgoing_edge_partitions_starting_at_vertex(
                pre_synaptic_vertex))

        # Partitions and Partition.edges will be OrderedSet but may be empty
        for partition in partitions:
            if partition.identifier != partition_id:
                continue
            for edge in partition.edges:
                if edge.post_vertex == post_synaptic_vertex:
                    return edge

        return None

    def _get_synaptic_data(
            self, as_list: bool, data_to_get: List[str],
            fixed_values: List[Tuple[str, int]],
            notify: Optional[Callable[[ConnectionHolder], None]]
            ) -> ConnectionHolder:
        post_vertex = self.__projection_edge.post_vertex
        pre_vertex = self.__projection_edge.pre_vertex

        # If in virtual board mode, the connection data should be set
        if self.__virtual_connection_list is not None:
            connection_holder = ConnectionHolder(
                data_to_get, as_list, pre_vertex.n_atoms, post_vertex.n_atoms,
                self.__virtual_connection_list, fixed_values=fixed_values,
                notify=notify)
            connection_holder.finish()
            return connection_holder

        # if not virtual board, make connection holder to be filled in at
        # possible later date
        connection_holder = ConnectionHolder(
            data_to_get, as_list, pre_vertex.n_atoms, post_vertex.n_atoms,
            fixed_values=fixed_values, notify=notify)

        # If we haven't run, add the holder to get connections, and return it
        # and set up a callback for after run to fill in this connection holder
        if not SpynnakerDataView.is_ran_ever():
            self.__synapse_information.add_pre_run_connection_holder(
                connection_holder)
            return connection_holder

        # Otherwise, get the connections now, as we have ran and therefore can
        # get them
        connections = post_vertex.get_connections_from_machine(
            self.__projection_edge, self.__synapse_information)
        if connections is not None:
            connection_holder.add_connections(connections)
            connection_holder.finish()
        return connection_holder

    def _clear_cache(self) -> None:
        post_vertex = self.__projection_edge.post_vertex
        if isinstance(post_vertex, AbstractAcceptsIncomingSynapses):
            post_vertex.clear_connection_cache()

    # -----------------------------------------------------------------

    def set(self, **attributes: Any) -> Never:  # @UnusedVariable
        """
        .. warning::
            Not implemented.
        """
        _ = (attributes)
        raise NotImplementedError("sPyNNaker does not currently do set")

    def size(self, gather: bool = True) -> Never:  # @UnusedVariable
        """
        Return the total number of connections.

        .. note::
            SpiNNaker always gathers.

        .. warning::
            Not implemented.

        :param gather:
            If False, only get the number of connections locally.
        """
        _ = gather
        raise NotImplementedError("sPyNNaker does not currently do size")

    def set_download_synapses(self, download_synapses: bool) -> None:
        """
        Set whether synapses should be downloaded when the simulation pauses.

        :param download_synapses: Whether to download synapses or not
        """
        self.__synapse_information.download_on_pause = download_synapses



# Copyright (c) 2022 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
import logging
from typing import (
    Any, cast, Dict, Iterable, List, Optional, overload, Sequence, Union,
    TYPE_CHECKING)

import numpy
from numpy import floating
from numpy.typing import NDArray
from pyNN.descriptions import TemplateEngine
import neo  # type: ignore[import]

from spinn_utilities.ranged.abstract_sized import AbstractSized, Selector
from spinn_utilities.log import FormatAdapter
from spinn_utilities.overrides import overrides

from spynnaker.pyNN.models.populations import Population
from spynnaker.pyNN.types import ViewIndices
from spynnaker.pyNN.utilities.neo_buffer_database import NeoBufferDatabase
from spynnaker.pyNN.utilities.utility_calls import get_neo_io
from spynnaker.pyNN.models.common.types import Names

if TYPE_CHECKING:
    from .neo_buffer_database import Annotations

logger = FormatAdapter(logging.getLogger(__file__))
_SELECTIVE_RECORDED_MSG = (
    "Getting data on a whole population when selective recording was "
    "active will result in only the recorded neurons being returned "
    "in numerical order and without repeats.")


class DataPopulation(object):
    """
    A wrapper of a sqlite3 database to provide the Population data methods
    """
    __slots__ = (
        "__database_file",
        "__label",
        "_indexes",
        "_size")

    def __init__(self, database_file: str, label: str,
                 indexes: ViewIndices = None):
        """
        :param database_file:
            The name of a file that contains an SQLite
            database holding the data.
        :param label: Label of the Population
        :param indexes: The indexes for which data should be returned.
            If ``None``, all data (view_index = data_indexes)
        """
        self.__label = label
        self.__database_file = database_file
        # getting size right away also check the inputs or fails fast
        with NeoBufferDatabase(self.__database_file) as db:
            size = db.get_population_metadata(label)[0]
        self._size = size
        self._indexes = indexes

    @overrides(Population.write_data)
    def write_data(self, io: Union[str, neo.baseio.BaseIO],
                   variables: Names = 'all', gather: bool = True,
                   clear: bool = False,
                   annotations: Annotations = None) -> None:
        # pylint: disable=missing-function-docstring,protected-access
        Population._check_params(gather, annotations)
        if clear:
            logger.warning("Ignoring clear as supported in this mode")
        if isinstance(io, str):
            io = get_neo_io(io)
        data = self.get_data(variables)
        # write the neo block to the file
        io.write(bl=data)

    @overrides(Population.describe)
    def describe(self, template: Optional[str] = None,
                 engine: Optional[Union[str, TemplateEngine]] = None
                 ) -> Union[str, Dict[str, Any]]:
        # pylint: disable=missing-function-docstring
        if template is not None:
            logger.warning("Ignoring template as not supported in this mode")
        if engine is not None:
            logger.warning("Ignoring engine as not supported in this mode")
        with NeoBufferDatabase(self.__database_file) as db:
            _, _, description = db.get_population_metadata(self.label)
            return description

    @overrides(Population.get_data)
    def get_data(
            self, variables: Names = 'all',
            gather: bool = True, clear: bool = False, *,
            annotations: Optional[Dict[str, Any]] = None) -> neo.Block:
        # pylint: disable=missing-function-docstring,protected-access
        Population._check_params(gather, annotations)
        if clear:
            logger.warning("Ignoring clear as supported in this mode")
        with NeoBufferDatabase(self.__database_file) as db:
            return db.get_full_block(
                self.__label, variables, self._indexes, annotations)

    @overrides(Population.spinnaker_get_data)
    def spinnaker_get_data(
            self, variable: str, as_matrix: bool = False,
            view_indexes: Optional[Sequence[int]] = None) -> NDArray[floating]:
        # pylint: disable=missing-function-docstring
        if view_indexes:
            return self[view_indexes].spinnaker_get_data(variable, as_matrix)
        with NeoBufferDatabase(self.__database_file) as db:
            return db.spinnaker_get_data(
                self.__label, variable, as_matrix, self._indexes)

    @overrides(Population.get_spike_counts)
    def get_spike_counts(self, gather: bool = True) -> Dict[int, int]:
        # pylint: disable=missing-function-docstring
        Population._check_params(gather)  # pylint: disable=protected-access
        with NeoBufferDatabase(self.__database_file) as db:
            return db.get_spike_counts(self.__label, self._indexes)

    @overrides(Population.find_units)
    def find_units(self, variable: str) -> Optional[str]:
        # pylint: disable=missing-function-docstring
        with NeoBufferDatabase(self.__database_file) as db:
            return db.find_units(self.__label, variable)

    def __len__(self) -> int:
        return self._size

    @property
    @overrides(Population.label)
    def label(self) -> str:
        # pylint: disable=missing-function-docstring
        return self.__label

    @property
    @overrides(Population.local_size)
    def local_size(self) -> int:
        # pylint: disable=missing-function-docstring
        return self._size

    @property
    @overrides(Population.size)
    def size(self) -> int:
        # pylint: disable=missing-function-docstring
        return self._size

    @overload
    def id_to_index(self, id: int) -> int:  # @ReservedAssignment
        # pylint: disable=redefined-builtin
        ...

    @overload
    def id_to_index(
            self, id: Iterable[int]) -> List[int]:  # @ReservedAssignment
        # pylint: disable=redefined-builtin
        ...

    @overrides(Population.id_to_index)
    def id_to_index(self, id: Union[int, Iterable[int]]
                    ) -> Union[int, List[int]]:  # @ReservedAssignment
        # pylint: disable=missing-function-docstring,redefined-builtin
        # assuming not called often so not caching first id
        with NeoBufferDatabase(self.__database_file) as db:
            _, first_id, _ = db.get_population_metadata(self.__label)
        last_id = self._size + first_id
        if not numpy.iterable(id):
            id = cast(int, id)
            if not first_id <= id <= last_id:
                raise ValueError(
                    f"id should be in the range [{first_id},{last_id}], "
                    f"actually {id}")
            return int(id - first_id)  # assume IDs are consecutive
        return [_id - first_id for _id in id]

    @overload
    def index_to_id(self, index: int) -> int:
        ...

    @overload
    def index_to_id(self, index: Iterable[int]) -> List[int]:
        ...

    @overrides(Population.index_to_id)
    def index_to_id(self, index: Union[int, Iterable[int]]
                    ) -> Union[int, List[int]]:
        # pylint: disable=missing-function-docstring
        # assuming not called often so not caching first id
        with NeoBufferDatabase(self.__database_file) as db:
            _, first_id, _ = db.get_population_metadata(self.__label)
        if not numpy.iterable(index):
            index = cast(int, index)
            if index >= self._size:
                raise ValueError(
                    f"indexes should be in the range [0,{self._size}],"
                    f" actually {index}")
            return int(index + first_id)
        # this assumes IDs are consecutive
        return [_index + first_id for _index in index]

    def __getitem__(self, index_or_slice: Selector) -> DataPopulation:
        """
        :param index_or_slice: a slice or numpy mask array.
            The mask array should either be a Boolean array (ideally) of the
            same size as the parent,
            or an integer array containing cell indices,
            i.e. if `p.size == 5` then:

            ::

                PopulationView(p, array([False, False, True, False, True]))
                PopulationView(p, array([2, 4]))
                PopulationView(p, slice(2, 5, 2))

            will all create the same view.
        :param index_or_slice:
        :return:
        """
        sized = AbstractSized(self._size)
        ids = sized.selector_to_ids(index_or_slice, warn=True)
        if self._indexes:
            indexes = [self._indexes[index] for index in ids]
        else:
            indexes = [range(self._size)[index] for index in ids]
        return DataPopulation(self.__database_file, self.__label, indexes)

    @overrides(Population.mean_spike_count)
    def mean_spike_count(self, gather: bool = True) -> float:
        # pylint: disable=missing-function-docstring
        Population._check_params(gather)  # pylint: disable=protected-access
        counts = self.get_spike_counts()
        return sum(counts.values()) / len(counts)



# Copyright (c) 2021 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
import logging
from typing import Iterator, Optional, Set, Tuple, TYPE_CHECKING

from spinn_utilities.log import FormatAdapter

from spinn_front_end_common.data import FecDataView

from spynnaker import _version

if TYPE_CHECKING:
    from spynnaker.pyNN.models.projection import Projection
    from spynnaker.pyNN.models.populations import Population

logger = FormatAdapter(logging.getLogger(__name__))
# pylint: disable=protected-access


class _SpynnakerDataModel(object):
    """
    Singleton data model.

    This class should not be accessed directly please use the DataView and
    DataWriter classes.
    Accessing or editing the data held here directly is *not supported!*

    There are other DataModel classes which sit next to this one and hold
    additional data. The DataView and DataWriter classes will combine these
    as needed.

    What data is held where and how can change without notice.
    """

    __singleton: Optional['_SpynnakerDataModel'] = None

    __slots__ = (
        # Data values cached
        "_id_counter",
        "_min_delay",
        "_populations",
        "_projections")

    def __new__(cls) -> '_SpynnakerDataModel':
        if cls.__singleton is not None:
            return cls.__singleton
        obj = object.__new__(cls)
        cls.__singleton = obj
        obj._clear()
        return obj

    def _clear(self) -> None:
        """
        Clears out all data.
        """
        self._id_counter = 0
        self._min_delay: Optional[float] = None
        # Using a dict to verify if later could be stored here only
        self._populations: Set[Population] = set()
        self._projections: Set[Projection] = set()

    def _hard_reset(self) -> None:
        """
        Puts all data back into the state expected at graph changed and
        `sim.reset`.
        """
        self._soft_reset()

    def _soft_reset(self) -> None:
        """
        Puts all data back into the state expected at `sim.reset` but not
        graph changed.
        """
        # segment_counter is increased by the writer


class SpynnakerDataView(FecDataView):
    """
    Adds the extra Methods to the View for PyNN level.

    See :py:class:`~spinn_utilities.data.UtilsDataView` for a more detailed
    description.

    Use this class wherever possible as it inherits all methods from all View
    classes.
    """

    __spy_data = _SpynnakerDataModel()

    __slots__ = ()

    @classmethod
    def get_min_delay(cls) -> float:
        """
        The minimum supported delay if available, in milliseconds.

        Typically `simulation_time_step_per_ms` but may be a positive multiple
        of it.

        :returns: the minimum delay in milliseconds.
        :raises ~spinn_utilities.exceptions.SpiNNUtilsException:
            If the min_delay is currently unavailable
        """
        if cls.__spy_data._min_delay is not None:
            return cls.__spy_data._min_delay
        return cls.get_simulation_time_step_ms()

    @classmethod
    def has_min_delay(cls) -> bool:
        """
        Report if there is a minimum supported delay available.

        If there is no delay time step could be used.

        :returns: True if get_min_delay method will work
        """
        if cls.__spy_data._min_delay is not None:
            return True
        return cls.has_time_step()

    @classmethod
    def iterate_projections(cls) -> Iterator[Projection]:
        """
        An iteration of the projections previously added.

        The iteration will be empty if no projections added.

        Note: This method is backed by a set so does not guarantee order

        :returns: Unordered iterator of projections.
        """
        return iter(cls.__spy_data._projections)

    @classmethod
    def get_n_projections(cls) -> int:
        """
        :returns: The number of projections previously added.
        """
        return len(cls.__spy_data._projections)

    @classmethod
    def add_projection(cls, projection: Projection) -> None:
        """
        Called by each projection to add itself to the list.

        Usage other than from `Projection.__init__` is not supported and likely
        to raise an exception

        :param projection: Projection to add
        :raises ~spinn_utilities.exceptions.SpiNNUtilsException:
            If projections should not be added in the current state
        """
        # UGLY but needed to avoid circular import
        # pylint: disable=import-outside-toplevel
        from spynnaker.pyNN.models.projection import Projection as Proj
        cls.check_user_can_act()
        if projection in cls.__spy_data._projections:
            raise NotImplementedError(
                "This method should only be called from the Projection init")
        if not isinstance(projection, Proj):
            raise TypeError("The projection must be a Projection")
        cls.__spy_data._projections.add(projection)

    @classmethod
    def iterate_populations(cls) -> Iterator[Population]:
        """
        An iteration of the populations previously added.

        The iteration will be empty if no populations added.

        Note: This method is backed by a set so does not guarantee order

        :returns: Unordered iterator of Populations
        """
        return iter(cls.__spy_data._populations)

    @classmethod
    def get_n_populations(cls) -> int:
        """
        :returns: The number of populations previously added.
        """
        return len(cls.__spy_data._populations)

    @classmethod
    def add_population(cls, population: Population) -> Tuple[int, int]:
        """
        Called by each population to add itself to the list.

        Usage other than from `Population.__init__` is not supported and likely
        to raise an exception

        Increments the all population ID counter by the size of the population.

        :param population: Population to add
        :return: The first and last global IDs for this Population
        :raises ~spinn_utilites.exceptions.SimulatorRunningException:
            If `sim.run` is currently running
        :raises ~spinn_utilites.exceptions.SimulatorNotSetupException:
            If called before `sim.setup`
        :raises ~spinn_utilites.exceptions.SimulatorShutdownException:
            If called after `sim.end`
        """
        # UGLY but needed to avoid circular import
        # pylint: disable=import-outside-toplevel
        from spynnaker.pyNN.models.populations.population import (
            Population as Pop)
        cls.check_user_can_act()
        if not isinstance(population, Pop):
            raise TypeError("The population must be a Population")
        if population in cls.__spy_data._populations:
            raise NotImplementedError(
                "This method should only be called from the Population init")
        first_id = cls.__spy_data._id_counter
        cls.__spy_data._id_counter += population.size
        cls.__spy_data._populations.add(population)
        return first_id, cls.__spy_data._id_counter-1

    @classmethod
    def get_sim_name(cls) -> str:
        """
        :returns: The name to be returned by `pyNN.spiNNaker.name`.
        """
        return _version.NAME



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import Optional
from spinn_front_end_common.utility_models import MultiCastCommand
from spinn_front_end_common.utilities.exceptions import ConfigurationException

# structure of command is KKKKKKKKKKKKKKKKKKKKK-IIIIIII-F-DDD
# K = ignored key at the top of the command
# I = instruction
# F = format
# D = device
_OFFSET_TO_IGNORED_KEY = 11
_OFFSET_TO_I = 4
_OFFSET_TO_F = 3
_OFFSET_TO_D = 0
_KEY_MASK = 0xFFFFF800
_I_MASK = 0x7F0
_F_MASK = 0x8
_D_MASK = 0x7


# UART masks and shifts
RETINA_UART_SHIFT = 3 + _OFFSET_TO_I
RETINA_WITHOUT_UART_MASK = 0x670
PUSH_BOT_LASER_LED_SPEAKER_FREQUENCY_UART_SHIFT = 1 + _OFFSET_TO_D
PUSH_BOT_MOTOR_WITHOUT_UART_MASK = 0x7C0
PUSH_BOT_MOTOR_UART_SHIFT = 0 + _OFFSET_TO_I


def _munich_key(instr_id: int, dim: int = 0, format_bit: int = 0) -> int:
    return ((instr_id << _OFFSET_TO_I) | (format_bit << _OFFSET_TO_F) |
            (dim << _OFFSET_TO_D))


def get_munich_i(key: int) -> int:
    """
    :param key:
    :returns: Get the instruction field from the key.
    """
    return key & _I_MASK


def get_munich_f(key: int) -> int:
    """
    :returns: The format field from the key.
    """
    return key & _F_MASK


def get_munich_d(key: int) -> int:
    """
    :returns: The device field from the key.
    """
    return key & _D_MASK


def get_retina_i(key: int) -> int:
    """
    :param key:
    :returns: The key with the UART mask.

    """
    return key & RETINA_WITHOUT_UART_MASK


def get_push_bot_laser_led_speaker_frequency_i(key: int) -> int:
    """
    :param key:
    :returns: The instruction field from the key with the I mask.
    """
    return get_munich_i(key)


def get_push_bot_motor_i(key: int) -> int:
    """
    :returns:
       The key without the universal asynchronous receiver/transmitter mask.
    """

    return key & PUSH_BOT_MOTOR_WITHOUT_UART_MASK


# Specific fields in the key
_SENSOR_OUTGOING_OFFSET_TO_D = 2
_SENSOR_OUTGOING_OFFSET_TO_I = 7

# Payload fields
_PAYLOAD_RETINA_PAYLOAD_OFFSET = 29
_PAYLOAD_RETINA_PAYLOAD_MASK = 0xE0000000
_PAYLOAD_RETINA_KEY_OFFSET = 26
_PAYLOAD_RETINA_KEY_MASK = 0x1C000000
_PAYLOAD_SENSOR_ID_OFFSET = 27
_PAYLOAD_OFFSET_FOR_SENSOR_TIME = 0


def GET_RETINA_KEY_VALUE(payload: int) -> int:
    # pylint: disable=invalid-name
    """
    :param payload:
    :returns: The payload with the retina key mask and offset.
    """
    return (payload & _PAYLOAD_RETINA_KEY_MASK) >> _PAYLOAD_RETINA_KEY_OFFSET


def GET_RETINA_PAYLOAD_VALUE(payload: int) -> int:
    # pylint: disable=invalid-name
    """
    :param payload:
    :returns: The payload with the retina payload mask and offset.
    """
    return (
        (payload & _PAYLOAD_RETINA_PAYLOAD_MASK) >>
        _PAYLOAD_RETINA_PAYLOAD_OFFSET
    )


#: command key for setting up the master key of the board
CONFIGURE_MASTER_KEY = _munich_key(127, 0)

#: command key for setting up what mode of device running on the board
CHANGE_MODE = _munich_key(127, 1)

#: command for turning off retina output
DISABLE_RETINA_EVENT_STREAMING = _munich_key(0, 0)

#: command for retina where payload is events
ACTIVE_RETINA_EVENT_STREAMING_KEYS_CONFIGURATION = _munich_key(0, 1)

#: command for retina where events are the key
ACTIVE_RETINA_EVENT_STREAMING_SET_KEY = _munich_key(0, 2)

#: set timer / counter for timestamps
SET_TIMER_COUNTER_FOR_TIMESTAMPS = _munich_key(0, 3)

#: handle master / slave time sync
MASTER_SLAVE_KEY = _munich_key(0, 4)

#: command for setting bias (whatever the check that is)
BIAS_KEY = _munich_key(0, 5)

#: reset retina key.
RESET_RETINA_KEY = _munich_key(0, 7)

#: request on-board sensor data
SENSOR_REPORTING_OFF_KEY = _munich_key(1, 0)

#: poll sensors once
POLL_SENSORS_ONCE_KEY = _munich_key(1, 1)

#: poll sensors continuously
POLL_SENSORS_CONTINUOUSLY_KEY = _munich_key(1, 2)

#: disable motor
ENABLE_DISABLE_MOTOR_KEY = _munich_key(2, 0)

#: run motor for total period
MOTOR_RUN_FOR_PERIOD_KEY = _munich_key(2, 1)

#: raw output for motor 0 (permanent)
MOTOR_0_RAW_PERM_KEY = _munich_key(2, 4)

#: raw output for motor 1 (permanent)
MOTOR_1_RAW_PERM_KEY = _munich_key(2, 5)

#: raw output for motor 0 (leak towards 0)
MOTOR_0_RAW_LEAK_KEY = _munich_key(2, 6)

#: raw output for motor 1 (leak towards 0)
MOTOR_1_RAW_LEAK_KEY = _munich_key(2, 7)

#: motor output duration timer period
MOTOR_TIMER_A_TOTAL_PERIOD_KEY = _munich_key(3, 0)
#: motor output duration timer period
MOTOR_TIMER_B_TOTAL_PERIOD_KEY = _munich_key(3, 2)
#: motor output duration timer period
MOTOR_TIMER_C_TOTAL_PERIOD_KEY = _munich_key(3, 4)

#: motor output ratio active period
MOTOR_TIMER_A_CHANNEL_0_ACTIVE_PERIOD_KEY = _munich_key(4, 0)
#: motor output ratio active period
MOTOR_TIMER_A_CHANNEL_1_ACTIVE_PERIOD_KEY = _munich_key(4, 1)
#: motor output ratio active period
MOTOR_TIMER_B_CHANNEL_0_ACTIVE_PERIOD_KEY = _munich_key(4, 2)
#: motor output ratio active period
MOTOR_TIMER_B_CHANNEL_1_ACTIVE_PERIOD_KEY = _munich_key(4, 3)
#: motor output ratio active period
MOTOR_TIMER_C_CHANNEL_0_ACTIVE_PERIOD_KEY = _munich_key(4, 4)
#: motor output ratio active period
MOTOR_TIMER_C_CHANNEL_1_ACTIVE_PERIOD_KEY = _munich_key(4, 5)

#: digital IO Signals
QUERY_STATES_LINES_KEY = _munich_key(5, 0)

#: set output pattern to payload
SET_OUTPUT_PATTERN_KEY = _munich_key(5, 1)

#: add payload (logic or (payload)) to current output
ADD_PAYLOAD_TO_CURRENT_OUTPUT_KEY = _munich_key(5, 2)

#: remove payload (logic or (payload)) to current output from current output
REMOVE_PAYLOAD_TO_CURRENT_OUTPUT_KEY = _munich_key(5, 3)

#: set payload pins to high impedance
SET_PAYLOAD_TO_HIGH_IMPEDANCE_KEY = _munich_key(5, 4)

#: set laser params for PushBot
PUSH_BOT_LASER_CONFIG_TOTAL_PERIOD = _munich_key(3, 0)
#: set laser params for PushBot
PUSH_BOT_LASER_CONFIG_ACTIVE_TIME = _munich_key(4, 0)
#: set laser params for PushBot
PUSH_BOT_LASER_FREQUENCY = _munich_key(37, 1)

#: set led params for PushBot
PUSH_BOT_LED_CONFIG_TOTAL_PERIOD = _munich_key(3, 4)
#: set led params for PushBot
PUSH_BOT_LED_BACK_CONFIG_ACTIVE_TIME = _munich_key(4, 4)
#: set led params for PushBot
PUSH_BOT_LED_FRONT_CONFIG_ACTIVE_TIME = _munich_key(4, 5)
#: set led params for PushBot
PUSH_BOT_LED_FREQUENCY = _munich_key(37, 0)

#: set speaker params for PushBot
PUSH_BOT_SPEAKER_CONFIG_TOTAL_PERIOD = _munich_key(3, 2)
#: set speaker params for PushBot
PUSH_BOT_SPEAKER_CONFIG_ACTIVE_TIME = _munich_key(4, 2)
#: set speaker params for PushBot
PUSH_BOT_SPEAKER_TONE_BEEP = _munich_key(36, 0)
#: set speaker params for PushBot
PUSH_BOT_SPEAKER_TONE_MELODY = _munich_key(36, 1)

#: PushBot motor control
PUSH_BOT_MOTOR_0_PERMANENT_VELOCITY = _munich_key(32, 0)
#: PushBot motor control
PUSH_BOT_MOTOR_1_PERMANENT_VELOCITY = _munich_key(32, 1)
#: PushBot motor control
PUSH_BOT_MOTOR_0_LEAKY_VELOCITY = _munich_key(32, 2)
#: PushBot motor control
PUSH_BOT_MOTOR_1_LEAKY_VELOCITY = _munich_key(32, 3)

# payload for master slave
_PAYLOAD_MASTER_SLAVE_USE_INTERNAL_COUNTER = 0
_PAYLOAD_MASTER_SLAVE_SET_SLAVE = 1
_PAYLOAD_MASTER_SLAVE_SET_MASTER_CLOCK_NOT_STARTED = 2
_PAYLOAD_MASTER_SLAVE_SET_MASTER_CLOCK_ACTIVE = 4


class RetinaKey(Enum):
    """
    The identification, pixels and bits per coordinate for each retina action.
    """
    FIXED_KEY = (0, 128, 7)
    NATIVE_128_X_128 = (1, 128, 7)
    DOWNSAMPLE_64_X_64 = (2, 64, 6)
    DOWNSAMPLE_32_X_32 = (3, 32, 5)
    DOWNSAMPLE_16_X_16 = (4, 16, 4)

    def __init__(self, ident: int, pixels: int, bits_per_coordinate: int):
        # pylint: disable=wrong-spelling-in-docstring
        """
        :param ident: The ID of the enum
        :param pixels: number of pixels per retina dimension
        :param bits_per_coordinate:  number of bits per retina dimension
        """
        self.__ident = ident << _PAYLOAD_RETINA_KEY_OFFSET
        self.__pixels = pixels
        self.__bits_per_coordinate = bits_per_coordinate

    @property
    def ident(self) -> int:
        # pylint: disable=wrong-spelling-in-docstring
        """
        The identification passed into the init.
        """
        return self.__ident

    @property
    def n_neurons(self) -> int:
        """
        The number of neurons passed into the init.
        """
        return 2 * (self.__pixels ** 2)

    @property
    def pixels(self) -> int:
        """
        The pixels passed into the init.
        """
        return self.__pixels

    @property
    def bits_per_coordinate(self) -> int:
        """
        The bits per coordinate passed into the init.
        """
        return self.__bits_per_coordinate


class RetinaPayload(Enum):
    """
    The indent and number of payload bytes for retina actions.
    """
    NO_PAYLOAD = (0, 0)
    EVENTS_IN_PAYLOAD = (0, 4)
    DELTA_TIMESTAMPS = (1, 4)
    ABSOLUTE_2_BYTE_TIMESTAMPS = (2, 2)
    ABSOLUTE_3_BYTE_TIMESTAMPS = (3, 3)
    ABSOLUTE_4_BYTE_TIMESTAMPS = (4, 4)

    def __init__(self, ident: int, n_payload_bytes: int):
        # pylint: disable=wrong-spelling-in-docstring
        """
        :param ident: ID for the enum
        :param n_payload_bytes: number of payload bytes for retina actions.
        """
        self.__ident = ident << _PAYLOAD_RETINA_PAYLOAD_OFFSET
        self.__n_payload_bytes = n_payload_bytes

    @property
    def ident(self) -> int:
        """
        The indent passed into the init.
        """
        return self.__ident

    @property
    def n_payload_bytes(self) -> int:
        """
        The n_payload_bytes passed into the init.
        """
        return self.__n_payload_bytes


class MUNICH_MODES(Enum):
    # pylint: disable=invalid-name
    """
    Types of modes supported by this protocol.
    """
    RESET_TO_DEFAULT = 0
    PUSH_BOT = 1
    SPOMNIBOT = 2
    BALL_BALANCER = 3
    MY_ORO_BOTICS = 4
    FREE = 5


class MunichIoSpiNNakerLinkProtocol(object):
    """
    Provides Multicast commands for the Munich SpiNNaker-Link protocol.
    """
    __slots__ = (
        "__instance_key",
        "__mode",
        "__uart_id")

    # The instance of the protocol in use, to ensure that each vertex that is
    # to send commands to the PushBot uses a different outgoing key; the top
    # part of the key is ignored, so this works out!
    _protocol_instance = 0

    # Keeps track of whether the mode has been configured already
    __sent_mode_command = False

    def __init__(self, mode: MUNICH_MODES, instance_key: Optional[int] = None,
                 uart_id: int = 0):
        """
        :param mode: The mode of operation of the protocol
        :param instance_key: The optional instance key to use
        :param uart_id: The ID of the UART when needed
        """
        self.__mode = mode

        # Create a key for this instance of the protocol
        # - see above for reasoning
        if instance_key is None:
            self.__instance_key = (
                MunichIoSpiNNakerLinkProtocol._protocol_instance <<
                _OFFSET_TO_IGNORED_KEY)
            MunichIoSpiNNakerLinkProtocol._protocol_instance += 1
        else:
            self.__instance_key = instance_key

        self.__uart_id = uart_id

    @property
    def mode(self) -> MUNICH_MODES:
        """
        spynnaker.pyNN.protocols.MUNICH_MODES
        """
        return self.__mode

    @property
    def uart_id(self) -> int:
        """
        The ID of the UART when needed
        """
        return self.__uart_id

    @property
    def instance_key(self) -> int:
        """
        The instance key to use
        """
        return self.__instance_key

    @staticmethod
    def sent_mode_command() -> bool:
        """
        :returns:
           True if the mode command has ever been requested by any instance.
        """
        return MunichIoSpiNNakerLinkProtocol.__sent_mode_command

    def _get_key(self, command: int,
                 offset_to_uart_id: Optional[int] = None) -> int:
        if offset_to_uart_id is None:
            return command | self.__instance_key
        return (
            command | self.__instance_key |
            (self.__uart_id << offset_to_uart_id))

    def configure_master_key(self, new_key: int,
                             time: Optional[int] = None) -> MultiCastCommand:
        """
        :param new_key:
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to configure master key.
        """
        return MultiCastCommand(
            self._get_key(CONFIGURE_MASTER_KEY), payload=new_key, time=time)

    def set_mode(self, time: Optional[int] = None) -> MultiCastCommand:
        """
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: The set mode. And records it was provided.
        """
        MunichIoSpiNNakerLinkProtocol.__sent_mode_command = True
        return MultiCastCommand(
            self._get_key(CHANGE_MODE), payload=self.__mode.value, time=time)

    @property
    def set_retina_key_key(self) -> int:
        """
        Key to set retina key.
        """
        return self._get_key(
            ACTIVE_RETINA_EVENT_STREAMING_SET_KEY, RETINA_UART_SHIFT)

    def set_retina_key(self, new_key: int,
                       time: Optional[int] = None) -> MultiCastCommand:
        """
        :param new_key:
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to set retina key.
        """
        return MultiCastCommand(
            self.set_retina_key_key, payload=new_key, time=time)

    @property
    def disable_retina_key(self) -> int:
        """
        Key to disable the retina.
        """
        return self._get_key(DISABLE_RETINA_EVENT_STREAMING, RETINA_UART_SHIFT)

    def disable_retina(self, time: Optional[int] = None) -> MultiCastCommand:
        """
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to disable retina.
        """
        return MultiCastCommand(self.disable_retina_key, time=time)

    def master_slave_use_internal_counter(
            self, time: Optional[int] = None) -> MultiCastCommand:
        """
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to set internal counter used.
        """
        return MultiCastCommand(
            self._get_key(MASTER_SLAVE_KEY, RETINA_UART_SHIFT),
            payload=_PAYLOAD_MASTER_SLAVE_USE_INTERNAL_COUNTER, time=time)

    def master_slave_set_slave(
            self, time: Optional[int] = None) -> MultiCastCommand:
        """
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to set slave.
        """
        return MultiCastCommand(
            self._get_key(MASTER_SLAVE_KEY, RETINA_UART_SHIFT),
            payload=_PAYLOAD_MASTER_SLAVE_SET_SLAVE, time=time)

    def master_slave_set_master_clock_not_started(
            self, time: Optional[int] = None) -> MultiCastCommand:
        """
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to set master clock active.
        """
        return MultiCastCommand(
            self._get_key(MASTER_SLAVE_KEY, RETINA_UART_SHIFT),
            payload=_PAYLOAD_MASTER_SLAVE_SET_MASTER_CLOCK_NOT_STARTED,
            time=time)

    def master_slave_set_master_clock_active(
            self, time: Optional[int] = None) -> MultiCastCommand:
        """
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to set master clock active.
        """
        return MultiCastCommand(
            self._get_key(MASTER_SLAVE_KEY, RETINA_UART_SHIFT),
            payload=_PAYLOAD_MASTER_SLAVE_SET_MASTER_CLOCK_ACTIVE,
            time=time)

    def bias_values(self, bias_id: int, bias_value: int,
                    time: Optional[int] = None) -> MultiCastCommand:
        """
        :param bias_id:
        :param bias_value:
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to bias level.
        """
        return MultiCastCommand(
            self._get_key(BIAS_KEY, RETINA_UART_SHIFT),
            payload=((bias_id << 0) | (bias_value << 8)), time=time)

    def reset_retina(self, time: Optional[int] = None) -> MultiCastCommand:
        """
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to reset retina.
        """
        return MultiCastCommand(
            self._get_key(RESET_RETINA_KEY, RETINA_UART_SHIFT), time=time)

    def turn_off_sensor_reporting(
            self, sensor_id: int,
            time: Optional[int] = None) -> MultiCastCommand:
        """
        :param sensor_id:
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to turn off sensor reporting.
        """
        return MultiCastCommand(
            self._get_key(SENSOR_REPORTING_OFF_KEY),
            payload=(sensor_id << _PAYLOAD_SENSOR_ID_OFFSET), time=time)

    def poll_sensors_once(self, sensor_id: int,
                          time: Optional[int] = None) -> MultiCastCommand:
        """
        :param sensor_id:
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to poll sensor once.
        """
        return MultiCastCommand(
            self._get_key(POLL_SENSORS_ONCE_KEY),
            payload=(sensor_id << _PAYLOAD_SENSOR_ID_OFFSET), time=time)

    def poll_individual_sensor_continuously(
            self, sensor_id: int, time_in_ms: int,
            time: Optional[int] = None) -> MultiCastCommand:
        """
        :param sensor_id:
        :param time_in_ms: time to sensor
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to sensor continuously.
        """
        return MultiCastCommand(
            self._get_key(POLL_SENSORS_CONTINUOUSLY_KEY),
            payload=((sensor_id << _PAYLOAD_SENSOR_ID_OFFSET) |
                     (time_in_ms << _PAYLOAD_OFFSET_FOR_SENSOR_TIME)),
            time=time)

    @property
    def enable_disable_motor_key(self) -> int:
        """
        Get key to disable motor.
        """
        return self._get_key(ENABLE_DISABLE_MOTOR_KEY, RETINA_UART_SHIFT)

    def generic_motor_enable(
            self, time: Optional[int] = None) -> MultiCastCommand:
        """
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to enable motor.
        """
        return MultiCastCommand(
            self.enable_disable_motor_key, payload=1, time=time)

    def generic_motor_disable(
            self, time: Optional[int] = None) -> MultiCastCommand:
        """
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to disable motor.
        """
        return MultiCastCommand(
            self.enable_disable_motor_key, payload=0, time=time)

    def generic_motor_total_period(
            self, time_in_ms: int,
            time: Optional[int] = None) -> MultiCastCommand:
        """
        :param time_in_ms:
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to set motor total period.
        """
        return MultiCastCommand(
            self._get_key(MOTOR_RUN_FOR_PERIOD_KEY, RETINA_UART_SHIFT),
            payload=time_in_ms, time=time)

    def generic_motor0_raw_output_permanent(
            self, pwm_signal: int,
            time: Optional[int] = None) -> MultiCastCommand:
        """
        :param pwm_signal:
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to set generic motor 0 raw output permanent.
        """
        return MultiCastCommand(
            self._get_key(MOTOR_0_RAW_PERM_KEY, RETINA_UART_SHIFT),
            payload=pwm_signal, time=time)

    def generic_motor1_raw_output_permanent(
            self, pwm_signal: int,
            time: Optional[int] = None) -> MultiCastCommand:
        """
        :param pwm_signal:
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to set generic motor 1 raw output permanent.
        """
        return MultiCastCommand(
            self._get_key(MOTOR_1_RAW_PERM_KEY, RETINA_UART_SHIFT),
            payload=pwm_signal, time=time)

    def generic_motor0_raw_output_leak_to_0(
            self, pwm_signal: int,
            time: Optional[int] = None) -> MultiCastCommand:
        """
        :param pwm_signal:
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to set generic motor 0 raw output leak to 0.
        """
        return MultiCastCommand(
            self._get_key(MOTOR_0_RAW_LEAK_KEY, RETINA_UART_SHIFT),
            payload=pwm_signal, time=time)

    def generic_motor1_raw_output_leak_to_0(
            self, pwm_signal: int,
            time: Optional[int] = None) -> MultiCastCommand:
        """
        :param pwm_signal:
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to set generic motor 1 raw output leak to 0.
        """
        return MultiCastCommand(
            self._get_key(MOTOR_1_RAW_LEAK_KEY, RETINA_UART_SHIFT),
            payload=pwm_signal, time=time)

    def pwm_pin_output_timer_a_duration(
            self, timer_period: int,
            time: Optional[int] = None) -> MultiCastCommand:
        """
        :param timer_period:
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to set a output timer duration.
        """
        return MultiCastCommand(
            self._get_key(MOTOR_TIMER_A_TOTAL_PERIOD_KEY, RETINA_UART_SHIFT),
            payload=timer_period, time=time)

    def pwm_pin_output_timer_b_duration(
            self, timer_period: int,
            time: Optional[int] = None) -> MultiCastCommand:
        """
        :param timer_period:
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to set b output timer duration.
        """
        return MultiCastCommand(
            self._get_key(MOTOR_TIMER_B_TOTAL_PERIOD_KEY, RETINA_UART_SHIFT),
            payload=timer_period, time=time)

    def pwm_pin_output_timer_c_duration(
            self, timer_period: int,
            time: Optional[int] = None) -> MultiCastCommand:
        """
        :param timer_period:
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to set c output timer duration.
        """
        return MultiCastCommand(
            self._get_key(MOTOR_TIMER_C_TOTAL_PERIOD_KEY, RETINA_UART_SHIFT),
            payload=timer_period, time=time)

    def pwm_pin_output_timer_a_channel_0_ratio(
            self, timer_period: int,
            time: Optional[int] = None) -> MultiCastCommand:
        """
        :param timer_period:
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to set a channel 0 output timer.
        """
        return MultiCastCommand(
            self._get_key(
                MOTOR_TIMER_A_CHANNEL_0_ACTIVE_PERIOD_KEY, RETINA_UART_SHIFT),
            payload=timer_period, time=time)

    def pwm_pin_output_timer_a_channel_1_ratio(
            self, timer_period: int,
            time: Optional[int] = None) -> MultiCastCommand:
        """
        :param timer_period:
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to set a channel 1 output timer.
        """
        return MultiCastCommand(
            self._get_key(
                MOTOR_TIMER_A_CHANNEL_1_ACTIVE_PERIOD_KEY, RETINA_UART_SHIFT),
            payload=timer_period, time=time)

    def pwm_pin_output_timer_b_channel_0_ratio(
            self, timer_period: int,
            time: Optional[int] = None) -> MultiCastCommand:
        """
        :param timer_period:
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to set b channel 0 output timer.

        """
        return MultiCastCommand(
            self._get_key(
                MOTOR_TIMER_B_CHANNEL_0_ACTIVE_PERIOD_KEY, RETINA_UART_SHIFT),
            payload=timer_period, time=time)

    def pwm_pin_output_timer_b_channel_1_ratio(
            self, timer_period: int,
            time: Optional[int] = None) -> MultiCastCommand:
        """
        :param timer_period:
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to set b channel 1 output timer.
        """
        return MultiCastCommand(
            self._get_key(
                MOTOR_TIMER_B_CHANNEL_1_ACTIVE_PERIOD_KEY, RETINA_UART_SHIFT),
            payload=timer_period, time=time)

    def pwm_pin_output_timer_c_channel_0_ratio(
            self, timer_period: int,
            time: Optional[int] = None) -> MultiCastCommand:
        """
        :param timer_period:
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to set c channel 0 output timer.
        """
        return MultiCastCommand(
            self._get_key(
                MOTOR_TIMER_C_CHANNEL_0_ACTIVE_PERIOD_KEY, RETINA_UART_SHIFT),
            payload=timer_period, time=time)

    def pwm_pin_output_timer_c_channel_1_ratio(
            self, timer_period: int,
            time: Optional[int] = None) -> MultiCastCommand:
        """
        :param timer_period:
        :returns: Command to set c channel 1 output timer.
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        """
        return MultiCastCommand(
            self._get_key(
                MOTOR_TIMER_C_CHANNEL_1_ACTIVE_PERIOD_KEY, RETINA_UART_SHIFT),
            payload=timer_period, time=time)

    def query_state_of_io_lines(
            self, time: Optional[int] = None) -> MultiCastCommand:
        """
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to query state of io lines.
        """
        return MultiCastCommand(
            self._get_key(QUERY_STATES_LINES_KEY), time=time)

    def set_output_pattern_for_payload(
            self, payload: int,
            time: Optional[int] = None) -> MultiCastCommand:
        """
        :param payload:
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to set output pattern for payload.
        """
        return MultiCastCommand(
            self._get_key(SET_OUTPUT_PATTERN_KEY), payload=payload, time=time)

    def add_payload_logic_to_current_output(
            self, payload: int,
            time: Optional[int] = None) -> MultiCastCommand:
        """
        :param payload:
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to add payload logic to current output.
        """
        return MultiCastCommand(
            self._get_key(ADD_PAYLOAD_TO_CURRENT_OUTPUT_KEY),
            payload=payload, time=time)

    def remove_payload_logic_to_current_output(
            self, payload: int,
            time: Optional[int] = None) -> MultiCastCommand:
        """
        :param payload:
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to remove payload logic to current output.
        """
        return MultiCastCommand(
            self._get_key(REMOVE_PAYLOAD_TO_CURRENT_OUTPUT_KEY),
            payload=payload, time=time)

    def set_payload_pins_to_high_impedance(
            self, payload: int,
            time: Optional[int] = None) -> MultiCastCommand:
        """
        :param payload:
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to set the payload pins to high impedance.
        """
        return MultiCastCommand(
            self._get_key(SET_PAYLOAD_TO_HIGH_IMPEDANCE_KEY),
            payload=payload, time=time)

    def _check_for_pushbot_mode(self) -> None:
        if self.__mode is not MUNICH_MODES.PUSH_BOT:
            raise ConfigurationException(
                "The mode you configured is not the PushBot, and so this "
                f"message is invalid for mode {self.__mode}")

    @property
    def push_bot_laser_config_total_period_key(self) -> int:
        """
        The key to set the laser total period.
        """
        return self._get_key(
            PUSH_BOT_LASER_CONFIG_TOTAL_PERIOD, RETINA_UART_SHIFT)

    def push_bot_laser_config_total_period(
            self, total_period: int,
            time: Optional[int] = None) -> MultiCastCommand:
        """
        :param total_period:
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to set the laser total period.
        :raises ConfigurationException: If the mode is not PUSH_BOT
        """
        self._check_for_pushbot_mode()
        return MultiCastCommand(
            self.push_bot_laser_config_total_period_key,
            payload=total_period, time=time)

    @property
    def push_bot_laser_config_active_time_key(self) -> int:
        """
        The key to set the laser active time.
        """
        return self._get_key(
            PUSH_BOT_LASER_CONFIG_ACTIVE_TIME, RETINA_UART_SHIFT)

    def push_bot_laser_config_active_time(
            self, active_time: int,
            time: Optional[int] = None) -> MultiCastCommand:
        """
        :param active_time: The time for the laser
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to set the laser active time.
        :raises ConfigurationException: If the mode is not PUSH_BOT
        """
        self._check_for_pushbot_mode()
        return MultiCastCommand(
            self.push_bot_laser_config_active_time_key,
            payload=active_time, time=time)

    @property
    def push_bot_laser_set_frequency_key(self) -> int:
        """
        The key to set the frequency.
        """
        return self._get_key(
            PUSH_BOT_LASER_FREQUENCY,
            PUSH_BOT_LASER_LED_SPEAKER_FREQUENCY_UART_SHIFT)

    def push_bot_laser_set_frequency(
            self, frequency: int,
            time: Optional[int] = None) -> MultiCastCommand:
        """
        :param frequency:
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to set the laser frequency.
        :raises ConfigurationException: If the mode is not PUSH_BOT
        """
        self._check_for_pushbot_mode()
        return MultiCastCommand(
            self.push_bot_laser_set_frequency_key,
            payload=frequency, time=time)

    @property
    def push_bot_speaker_config_total_period_key(self) -> int:
        """
        The key to set the speaker total period.
        """
        return self._get_key(
            PUSH_BOT_SPEAKER_CONFIG_TOTAL_PERIOD, RETINA_UART_SHIFT)

    def push_bot_speaker_config_total_period(
            self, total_period: int,
            time: Optional[int] = None) -> MultiCastCommand:
        """
        :param total_period:
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to set the speaker total period.
        :raises ConfigurationException: If the mode is not PUSH_BOT
        """
        self._check_for_pushbot_mode()
        return MultiCastCommand(
            self.push_bot_speaker_config_total_period_key,
            payload=total_period, time=time)

    @property
    def push_bot_speaker_config_active_time_key(self) -> int:
        """
        The key to set the speaker active time.
        """
        return self._get_key(
            PUSH_BOT_SPEAKER_CONFIG_ACTIVE_TIME, RETINA_UART_SHIFT)

    def push_bot_speaker_config_active_time(
            self, active_time: int,
            time: Optional[int] = None) -> MultiCastCommand:
        """
        :param active_time:
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to set the speaker active time.
        :raises ConfigurationException: If the mode is not PUSH_BOT
        """
        self._check_for_pushbot_mode()
        return MultiCastCommand(
            self.push_bot_speaker_config_active_time_key,
            payload=active_time, time=time)

    @property
    def push_bot_speaker_set_tone_key(self) -> int:
        """
        The key to set the tone.
        """
        return self._get_key(
            PUSH_BOT_SPEAKER_TONE_BEEP,
            PUSH_BOT_LASER_LED_SPEAKER_FREQUENCY_UART_SHIFT)

    def push_bot_speaker_set_tone(
            self, frequency: int,
            time: Optional[int] = None) -> MultiCastCommand:
        """
        :param frequency:
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to set the tone.
        :raises ConfigurationException: If the mode is not PUSH_BOT
        """
        self._check_for_pushbot_mode()
        return MultiCastCommand(
            self.push_bot_speaker_set_tone_key,
            payload=frequency, time=time)

    @property
    def push_bot_speaker_set_melody_key(self) -> int:
        """
        The key to set the melody.
        """
        return self._get_key(
            PUSH_BOT_SPEAKER_TONE_MELODY,
            PUSH_BOT_LASER_LED_SPEAKER_FREQUENCY_UART_SHIFT)

    def push_bot_speaker_set_melody(
            self, melody: int,
            time: Optional[int] = None) -> MultiCastCommand:
        """
        :param melody:
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to set the melody.
        :raises ConfigurationException: If the mode is not PUSH_BOT
        """
        self._check_for_pushbot_mode()
        return MultiCastCommand(
            self.push_bot_speaker_set_melody_key,
            payload=melody, time=time)

    @property
    def push_bot_led_total_period_key(self) -> int:
        """
        The key to set the total led period.
        """
        return self._get_key(
            PUSH_BOT_LED_CONFIG_TOTAL_PERIOD, RETINA_UART_SHIFT)

    def push_bot_led_total_period(
            self, total_period: int,
            time: Optional[int] = None) -> MultiCastCommand:
        """
        :param total_period: total period for the LED
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to set the total led period.
        :raises ConfigurationException: If the mode is not PUSH_BOT
        """
        self._check_for_pushbot_mode()
        return MultiCastCommand(
            self.push_bot_led_total_period_key,
            payload=total_period, time=time)

    @property
    def push_bot_led_back_active_time_key(self) -> int:
        """
        The key to set the back led active time.
        """
        return self._get_key(
            PUSH_BOT_LED_BACK_CONFIG_ACTIVE_TIME, RETINA_UART_SHIFT)

    def push_bot_led_back_active_time(
            self, active_time: int,
            time: Optional[int] = None) -> MultiCastCommand:
        """
        :param active_time:
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to set the back led active time.
        :raises ConfigurationException: If the mode is not PUSH_BOT
        """
        self._check_for_pushbot_mode()
        return MultiCastCommand(
            self.push_bot_led_back_active_time_key,
            payload=active_time, time=time)

    @property
    def push_bot_led_front_active_time_key(self) -> int:
        """
        The key to set the front led active time.
        """
        return self._get_key(
            PUSH_BOT_LED_FRONT_CONFIG_ACTIVE_TIME, RETINA_UART_SHIFT)

    def push_bot_led_front_active_time(
            self, active_time: int,
            time: Optional[int] = None) -> MultiCastCommand:
        """
        :param active_time:
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to set the front led active time.
        :raises ConfigurationException: If the mode is not PUSH_BOT
        """
        self._check_for_pushbot_mode()
        return MultiCastCommand(
            self.push_bot_led_front_active_time_key,
            payload=active_time, time=time)

    @property
    def push_bot_led_set_frequency_key(self) -> int:
        """
        The key to set the led frequency.
        """
        return self._get_key(
            PUSH_BOT_LED_FREQUENCY,
            PUSH_BOT_LASER_LED_SPEAKER_FREQUENCY_UART_SHIFT)

    def push_bot_led_set_frequency(
            self, frequency: int,
            time: Optional[int] = None) -> MultiCastCommand:
        """
        :param frequency:
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to set the led frequency.
        :raises ConfigurationException: If the mode is not PUSH_BOT
        """
        self._check_for_pushbot_mode()
        return MultiCastCommand(
            self.push_bot_led_set_frequency_key,
            payload=frequency, time=time)

    @property
    def push_bot_motor_0_permanent_key(self) -> int:
        """
        The key for the change motor 0 permanently.
        """
        return self._get_key(
            PUSH_BOT_MOTOR_0_PERMANENT_VELOCITY, PUSH_BOT_MOTOR_UART_SHIFT)

    def push_bot_motor_0_permanent(
            self, velocity: int,
            time: Optional[int] = None) -> MultiCastCommand:
        """
        :param velocity:
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to change motor 0 permanently.
        :raises ConfigurationException: If the mode is not PUSH_BOT
        """
        self._check_for_pushbot_mode()
        return MultiCastCommand(
            self.push_bot_motor_0_permanent_key,
            payload=velocity, time=time)

    @property
    def push_bot_motor_1_permanent_key(self) -> int:
        """
        The key for the change motor 1 permanently.
        """
        return self._get_key(
            PUSH_BOT_MOTOR_1_PERMANENT_VELOCITY, PUSH_BOT_MOTOR_UART_SHIFT)

    def push_bot_motor_1_permanent(
            self, velocity: int,
            time: Optional[int] = None) -> MultiCastCommand:
        """
        :param velocity:
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to change motor 1 permanently.
        :raises ConfigurationException: If the mode is not PUSH_BOT
        """
        self._check_for_pushbot_mode()
        return MultiCastCommand(
            self.push_bot_motor_1_permanent_key,
            payload=velocity, time=time)

    @property
    def push_bot_motor_0_leaking_towards_zero_key(self) -> int:
        """
        The key for the change motor 0 towards zero.
        """
        return self._get_key(
            PUSH_BOT_MOTOR_0_LEAKY_VELOCITY, PUSH_BOT_MOTOR_UART_SHIFT)

    def push_bot_motor_0_leaking_towards_zero(
            self, velocity: int,
            time: Optional[int] = None) -> MultiCastCommand:
        """
        :param velocity:
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to change motor 0 towards zero.
        :raises ConfigurationException: If the mode is not PUSH_BOT
        """
        self._check_for_pushbot_mode()
        return MultiCastCommand(
            self.push_bot_motor_0_leaking_towards_zero_key,
            payload=velocity, time=time)

    @property
    def push_bot_motor_1_leaking_towards_zero_key(self) -> int:
        """
        The key for the change motor 1 towards zero.
        """
        return self._get_key(
            PUSH_BOT_MOTOR_1_LEAKY_VELOCITY, PUSH_BOT_MOTOR_UART_SHIFT)

    def push_bot_motor_1_leaking_towards_zero(
            self, velocity: int,
            time: Optional[int] = None) -> MultiCastCommand:
        """
        :param velocity:
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :returns: Command to change motor 1 towards zero.
        :raises ConfigurationException: If the mode is not PUSH_BOT
        """
        self._check_for_pushbot_mode()
        return MultiCastCommand(
            self.push_bot_motor_1_leaking_towards_zero_key,
            payload=velocity, time=time)

    def sensor_transmission_key(self, sensor_id: int) -> int:
        """
        :param sensor_id:
        :returns: The transmission key to this sensor id.
        """
        return ((sensor_id << _SENSOR_OUTGOING_OFFSET_TO_D) |
                (self.__uart_id << _SENSOR_OUTGOING_OFFSET_TO_I))

    @property
    def set_retina_transmission_key(self) -> int:
        """
        The key to set the retina_transmission.
        """
        return self._get_key(
            ACTIVE_RETINA_EVENT_STREAMING_KEYS_CONFIGURATION,
            RETINA_UART_SHIFT)

    def set_retina_transmission(
            self, retina_key: Optional[RetinaKey] = RetinaKey.NATIVE_128_X_128,
            retina_payload: Optional[RetinaPayload] = None,
            time: Optional[int] = None) -> MultiCastCommand:
        """
        Set the retina transmission key.

        :param retina_key: the new key for the retina
        :param retina_payload:
            the new payload for the set retina key command packet
        :param time: when to transmit this packet
        :return: the command to send
        """
        retina_key_id = retina_key.ident if retina_key is not None else 0

        if retina_payload is None:
            if retina_key == RetinaKey.FIXED_KEY:
                retina_payload = RetinaPayload.EVENTS_IN_PAYLOAD
            else:
                retina_payload = RetinaPayload.NO_PAYLOAD

        if (retina_key == RetinaKey.FIXED_KEY and
                retina_payload != RetinaPayload.EVENTS_IN_PAYLOAD):
            raise ConfigurationException(
                "If the Retina Key is FIXED_KEY, the payload must be"
                " EVENTS_IN_PAYLOAD")

        return MultiCastCommand(
            self.set_retina_transmission_key,
            payload=retina_key_id | retina_payload.ident,
            time=time)



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional, Sequence

import numpy as np
from numpy.typing import NDArray
from neo import AnalogSignal, Block, SpikeTrain
import quantities
from quantities import UnitTime

# needed as dealing with quantities


def convert_analog_signal(
        signal_array: AnalogSignal,
        time_unit: UnitTime = quantities.ms) -> NDArray:
    """
    Converts part of a NEO object into told spynnaker7 format.

    :param signal_array: Extended Quantities object
    :param time_unit:
        Data time unit for time index
    :returns: Data in Spynnaker (7) format
    """
    ids = signal_array.annotations["channel_names"]
    xs = range(len(ids))
    if time_unit == quantities.ms:
        times = signal_array.times.magnitude
    else:
        times = signal_array.times.rescale(time_unit).magnitude
    all_times = np.tile(times, len(xs))
    neurons = np.repeat(ids, len(times))
    values = np.concatenate([signal_array.magnitude[:, x] for x in xs])
    return np.column_stack((neurons, all_times, values))


def convert_data(data: Block, name: str, run: int = 0) -> NDArray:
    """
    Converts the data into a numpy array in the format ID, time, value.

    :param data: Data as returned by a getData() call
    :param name: Name of the data to be extracted.
        Same values as used in getData()
    :param run: Zero based index of the run to extract data for
    :returns: Data for the named data type in Spynnaker (7) format
    """
    if len(data.segments) <= run:
        raise ValueError(
            f"Data only contains {len(data.segments)} so unable to run {run}. "
            "Note run is the zero based index.")
    if name == "all":
        raise ValueError("Unable to convert all data in one go "
                         "as result would be comparing apples and oranges.")
    if name == "spikes":
        return convert_spikes(data, run)
    return convert_analog_signal(
        data.segments[run].filter(name=name)[0])


def convert_data_list(data: Block, name: str,
                      runs: Optional[Sequence[int]] = None) -> List[NDArray]:
    """
    Converts the data into a list of numpy arrays in the format ID, time,
    value.

    :param data: Data as returned by a getData() call
    :param name: Name of the data to be extracted.
        Same values as used in getData()
    :param runs: List of Zero based index of the run to extract data for.
        Or `None` to extract all runs
    :returns: List of numpy arrays for the named data in Spynnaker (7) format
    """
    if runs is None:
        runs = range(len(data.segments))
    return [
        convert_data(data, name, run=run)
        for run in runs]


def convert_v_list(
        data: Block, runs: Optional[Sequence[int]] = None) -> List[NDArray]:
    """
    Converts the voltage into a list numpy array one per segment (all
    runs) in the format ID, time, value.

    :param data: The data to convert; it must have V data in it
    :param runs: List of Zero based index of the run to extract data for.
        Or `None` to extract all runs
    :returns: Voltage in sPyNNaker (7) format
    """
    return convert_data_list(data, "v", runs=runs)


def convert_gsyn_exc_list(
        data: Block, runs: Optional[Sequence[int]] = None) -> List[NDArray]:
    """
    Converts the gsyn_exc into a list numpy array one per segment (all
    runs) in the format ID, time, value.

    :param data:
        The data to convert; it must have Gsyn_exc data in it
    :param runs: List of Zero based index of the run to extract data for.
        Or `None` to extract all runs
    :returns: Gsyn in sPyNNaker (7) format
    """
    return convert_data_list(data, "gsyn_exc", runs=runs)


def convert_gsyn_inh_list(
        data: Block, runs: Optional[Sequence[int]] = None) -> List[NDArray]:
    """
    Converts the gsyn_inh into a list numpy array one per segment (all
    runs) in the format ID, time, value.

    :param data:
        The data to convert; it must have Gsyn_inh data in it
    :param runs: List of Zero based index of the run to extract data for.
        Or `None` to extract all runs
    :returns: Gsyn in sPyNNaker (7) format
    """
    return convert_data_list(data, "gsyn_inh", runs=runs)


def convert_gsyn(gsyn_exc: Block, gsyn_inh: Block) -> NDArray:
    """
    Converts two neo objects into the spynnaker7 format.

    .. note::
        It is acceptable for both neo parameters to be the same object

    :param gsyn_exc: neo with gsyn_exc data
    :param gsyn_inh: neo with gsyn_inh data
    :returns: Gsyn in sPyNNaker (7) format
    """
    exc = gsyn_exc.segments[0].filter(name='gsyn_exc')[0]
    inh = gsyn_inh.segments[0].filter(name='gsyn_inh')[0]
    ids = exc.annotations["channel_names"]
    ids2 = inh.annotations["channel_names"]
    if len(ids) != len(ids2):
        raise ValueError(
            f"Found {len(ids)} neuron IDs in gsyn_exc "
            f"but {len(ids2)} in gsyn_inh")
    if not np.allclose(ids, ids2):
        raise ValueError("IDs in gsyn_exc and gsyn_inh do not match")
    times = exc.times.rescale(quantities.ms)
    times2 = inh.times.rescale(quantities.ms)
    if len(times) != len(times2):
        raise ValueError(
            f"Found {len(times)} times in gsyn_exc "
            f"but {len(times2)} in gsyn_inh")
    if not np.allclose(times, times2):
        raise ValueError("times in gsyn_exc and gsyn_inh do not match")
    all_times = np.tile(times, len(ids))
    neurons = np.repeat(ids, len(times))
    exc_np = np.concatenate([exc[:, x] for x in range(len(ids))])
    inh_np = np.concatenate([inh[:, x] for x in range(len(ids))])
    return np.column_stack((neurons, all_times, exc_np, inh_np))


def convert_spiketrains(spiketrains: List[SpikeTrain]) -> NDArray:
    """
    Converts a list of spiketrains into spynnaker7 format.

    :param spiketrains: List of SpikeTrains
    :returns: Spikes in sPyNNaker (7) format
    """
    if len(spiketrains) == 0:
        return np.empty(shape=(0, 2))

    neurons = np.concatenate([
        np.repeat(x.annotations['source_index'], len(x))
        for x in spiketrains])
    spikes = np.concatenate([x.magnitude for x in spiketrains])
    return np.column_stack((neurons, spikes))


def convert_spikes(neo: Block, run: int = 0) -> NDArray:
    """
    Extracts the spikes for run one from a Neo Object.

    :param neo: neo Object including Spike Data
    :param run: Zero based index of the run to extract data for
    :returns: Spikes in sPyNNaker (7) format
    """
    if len(neo.segments) <= run:
        raise ValueError(
            f"Data only contains {len(neo.segments)} so unable to run {run}. "
            "Note run is the zero based index.")
    return convert_spiketrains(neo.segments[run].spiketrains)


def count_spiketrains(spiketrains: SpikeTrain) -> int:
    """
    Help function to count the number of spikes in a list of spiketrains.

    :param spiketrains: List of SpikeTrains
    :return: Total number of spikes in all the spiketrains
    """
    return sum(map(len, spiketrains))


def count_spikes(neo: Block) -> int:
    """
    Help function to count the number of spikes in a list of spiketrains.

    Only counts run 0

    :param neo: Neo Object which has spikes in it
    :return: The number of spikes in the first segment
    """
    return count_spiketrains(neo.segments[0].spiketrains)



# Copyright (c) 2021 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from typing import TYPE_CHECKING
from spinn_utilities.abstract_base import AbstractBase, abstractmethod
if TYPE_CHECKING:
    from spynnaker.pyNN.models.projection import Projection


class AbstractSupportsSignedWeights(object, metaclass=AbstractBase):
    """
    A synapse dynamics object that supports signed weights.
    """

    @abstractmethod
    def get_positive_synapse_index(
            self, incoming_projection: Projection) -> int:
        """
        :param incoming_projection: The projection targeted
        :returns: The synapse type that positive weights will arrive as.
        """
        raise NotImplementedError

    @abstractmethod
    def get_negative_synapse_index(
            self, incoming_projection: Projection) -> int:
        """
        :param incoming_projection: The projection targeted
        :returns: The synapse type that negative weights will arrive as.
        """
        raise NotImplementedError

    @abstractmethod
    def get_maximum_positive_weight(
            self, incoming_projection: Projection) -> float:
        """
        Get the maximum likely positive weight.

        .. note::
            This must be a value >= 0.

        :param incoming_projection: The projection targeted
        :returns: The maximum likely positive weight.
        """
        raise NotImplementedError

    @abstractmethod
    def get_minimum_negative_weight(
            self, incoming_projection: Projection) -> float:
        """
        Get the minimum likely negative weight.

        .. note::
            This must be a value <= 0.

        :param incoming_projection: The projection targeted
        :returns: The minimum likely negative weight
        """
        raise NotImplementedError

    @abstractmethod
    def get_mean_positive_weight(
            self, incoming_projection: Projection) -> float:
        """
        Get the mean of the positive weights.

        .. note::
            This must be a value >= 0.

        :param incoming_projection: The projection targeted
        :returns: The mean of the positive weights.
        """
        raise NotImplementedError

    @abstractmethod
    def get_mean_negative_weight(
            self, incoming_projection: Projection) -> float:
        """
        Get the mean of the negative weights.

        .. note::
            This must be a value <= 0.

        :param incoming_projection: The projection targeted
        :returns: The mean of the negative weights.
        """
        raise NotImplementedError

    @abstractmethod
    def get_variance_positive_weight(
            self, incoming_projection: Projection) -> float:
        """
        Get the variance of the positive weights.

        .. note::
            This must be a value >= 0.

        :param incoming_projection: The projection targeted
        :returns: The variance of the positive weights.
        """
        raise NotImplementedError

    @abstractmethod
    def get_variance_negative_weight(
            self, incoming_projection: Projection) -> float:
        """
        Get the variance of the negative weights.

        .. note::
            This must be a value <= 0.

        :param incoming_projection: The projection targeted
        :returns: The variance of the negative weights.
        """
        raise NotImplementedError



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spynnaker.pyNN.models.neuron import AbstractPyNNNeuronModelStandard
from spynnaker.pyNN.models.defaults import default_initial_values
from spynnaker.pyNN.models.neuron.implementations import ModelParameter
from spynnaker.pyNN.models.neuron.neuron_models import (
    NeuronModelLeakyIntegrateAndFire)
from spynnaker.pyNN.models.neuron.synapse_types import SynapseTypeExponential
from spynnaker.pyNN.models.neuron.input_types import InputTypeCurrent
from spynnaker.pyNN.models.neuron.threshold_types import ThresholdTypeStatic
from spynnaker.pyNN.models.neuron.additional_inputs import (
    AdditionalInputCa2Adaptive)


class IFCurrExpCa2Adaptive(AbstractPyNNNeuronModelStandard):
    # pylint: disable=wrong-spelling-in-docstring
    """
    Model from Liu, Y. H., & Wang, X. J. (2001). Spike-frequency
    adaptation of a generalized leaky integrate-and-fire model neuron.
    *Journal of Computational Neuroscience*, 10(1), 25-45.
    `doi:10.1023/A:1008916026143
    <https://doi.org/10.1023/A:1008916026143>`_
    """

    @default_initial_values({"v", "isyn_exc", "isyn_inh", "i_ca2"})
    def __init__(
            self, tau_m: ModelParameter = 20.0, cm: ModelParameter = 1.0,
            v_rest: ModelParameter = -65.0, v_reset: ModelParameter = -65.0,
            v_thresh: ModelParameter = -50.0, tau_syn_E: ModelParameter = 5.0,
            tau_syn_I: ModelParameter = 5.0, tau_refrac: ModelParameter = 0.1,
            i_offset: ModelParameter = 0.0, tau_ca2: ModelParameter = 50.0,
            i_ca2: ModelParameter = 0.0, i_alpha: ModelParameter = 0.1,
            v: ModelParameter = -65.0, isyn_exc: ModelParameter = 0.0,
            isyn_inh: ModelParameter = 0.0):
        """
            :param tau_m: :math:`\\tau_m`
        :param cm: :math:`C_m`
        :param v_rest: :math:`V_{rest}`
        :param v_reset: :math:`V_{reset}`
        :param v_thresh: :math:`V_{thresh}`
        :param tau_syn_E: :math:`\\tau^{syn}_e`
        :param tau_syn_I: :math:`\\tau^{syn}_i`
        :param tau_refrac: :math:`\\tau_{refrac}`
        :param i_offset: :math:`I_{offset}`
        :param tau_ca2: :math:`\\tau_{\\mathrm{Ca}^{+2}}`
        :param i_ca2: :math:`I_{\\mathrm{Ca}^{+2}}`
        :param i_alpha: :math:`\\tau_\\alpha`
        :param v: :math:`V_{init}`
        :param isyn_exc: :math:`I^{syn}_e`
        :param isyn_inh: :math:`I^{syn}_i`
        """
        neuron_model = NeuronModelLeakyIntegrateAndFire(
            v, v_rest, tau_m, cm, i_offset, v_reset, tau_refrac)
        synapse_type = SynapseTypeExponential(
            tau_syn_E, tau_syn_I, isyn_exc, isyn_inh)
        input_type = InputTypeCurrent()
        threshold_type = ThresholdTypeStatic(v_thresh)
        additional_input_type = AdditionalInputCa2Adaptive(
            tau_ca2, i_ca2, i_alpha)

        super().__init__(
            model_name="IF_curr_exp_ca2_adaptive",
            binary="IF_curr_exp_ca2_adaptive.aplx",
            neuron_model=neuron_model, input_type=input_type,
            synapse_type=synapse_type, threshold_type=threshold_type,
            additional_input_type=additional_input_type)



# Copyright (c) 2016 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from collections.abc import Iterable
from typing import List, Optional, Union, cast, TYPE_CHECKING

from spinn_utilities.overrides import overrides
from pacman.model.graphs.application import ApplicationEdge

if TYPE_CHECKING:
    from spynnaker.pyNN.models.neuron import PopulationVertex
    from spynnaker.pyNN.models.utility_models.delays import (
        DelayExtensionVertex)
    from spynnaker.pyNN.models.neural_projections import (
        SynapseInformation, ProjectionApplicationEdge)


class DelayedApplicationEdge(ApplicationEdge):
    """
    The Edge from a delay vertex to a Population vertex.
    """
    __slots__ = (
        "__synapse_information",
        "__undelayed_edge")

    def __init__(
            self, pre_vertex: DelayExtensionVertex,
            post_vertex: PopulationVertex,
            synapse_information: Union[
                SynapseInformation, Iterable[SynapseInformation]],
            undelayed_edge: ProjectionApplicationEdge,
            label: Optional[str] = None):
        """
        :param pre_vertex: The delay extension at the start of the edge
        :param post_vertex: The target of the synapses
        :param synapse_information:
            The synapse information on this edge
        :param undelayed_edge:
            The edge that is used for projections without extended delays
        :param label: The edge label
        """
        super().__init__(pre_vertex, post_vertex, label=label)
        if isinstance(synapse_information, Iterable):
            self.__synapse_information = list(synapse_information)
        else:
            self.__synapse_information = [synapse_information]
        self.__undelayed_edge = undelayed_edge

    @property
    @overrides(ApplicationEdge.pre_vertex)
    def pre_vertex(self) -> DelayExtensionVertex:
        return cast('DelayExtensionVertex', super().pre_vertex)

    @property
    @overrides(ApplicationEdge.post_vertex)
    def post_vertex(self) -> PopulationVertex:
        return cast('PopulationVertex', super().post_vertex)

    @property
    def synapse_information(self) -> List[SynapseInformation]:
        """
        The synapse information on this edge
        """
        return self.__synapse_information

    def add_synapse_information(
            self, synapse_information: SynapseInformation) -> None:
        """
        Add more synapse information on this edge
        """
        self.__synapse_information.append(synapse_information)

    @property
    def undelayed_edge(self) -> ProjectionApplicationEdge:
        """
        The edge for projections without extended delays.
        """
        return self.__undelayed_edge



# Copyright (c) 2022 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import sys
from typing import cast, TYPE_CHECKING
from spinn_utilities.overrides import overrides
from pacman.utilities.utility_calls import get_keys
from spinn_front_end_common.utility_models import (
    ReverseIPTagMulticastSourceMachineVertex)
from spynnaker.pyNN.data.spynnaker_data_view import SpynnakerDataView
if TYPE_CHECKING:
    from .spike_source_array_vertex import SpikeSourceArrayVertex


class SpikeSourceArrayMachineVertex(ReverseIPTagMulticastSourceMachineVertex):
    """
    Extended to add colour.
    """

    @property
    def _pop_vertex(self) -> SpikeSourceArrayVertex:
        return cast('SpikeSourceArrayVertex', self.app_vertex)

    @overrides(
        ReverseIPTagMulticastSourceMachineVertex.get_n_keys_for_partition)
    def get_n_keys_for_partition(self, partition_id: str) -> int:
        n_keys = super().get_n_keys_for_partition(partition_id)
        n_colours = 2 ** self._pop_vertex.n_colour_bits
        return n_keys * n_colours

    @overrides(ReverseIPTagMulticastSourceMachineVertex._fill_send_buffer_1d)
    def _fill_send_buffer_1d(self, key_base: int) -> None:
        first_time_step = SpynnakerDataView.get_first_machine_time_step()
        end_time_step = (
                SpynnakerDataView.get_current_run_timesteps() or sys.maxsize)
        if first_time_step == end_time_step:
            return
        if self._send_buffer_times is None or self._send_buffer is None:
            return
        keys = get_keys(
            key_base, self.vertex_slice, self._pop_vertex.n_colour_bits)
        colour_mask = (2 ** self._pop_vertex.n_colour_bits) - 1
        for tick in sorted(self._send_buffer_times):
            if first_time_step <= tick < end_time_step:
                self._send_buffer.add_keys(
                    tick, keys + (tick & colour_mask))

    @overrides(ReverseIPTagMulticastSourceMachineVertex._fill_send_buffer_2d)
    def _fill_send_buffer_2d(self, key_base: int) -> None:
        first_time_step = SpynnakerDataView.get_first_machine_time_step()
        end_time_step = (
                SpynnakerDataView.get_current_run_timesteps() or sys.maxsize)
        if first_time_step == end_time_step:
            return
        if self._send_buffer_times is None or self._send_buffer is None:
            return
        keys = get_keys(
            key_base, self.vertex_slice, self._pop_vertex.n_colour_bits)
        colour_mask = (2 ** self._pop_vertex.n_colour_bits) - 1
        for atom in range(self.vertex_slice.n_atoms):
            for tick in sorted(self._send_buffer_times[atom]):
                if first_time_step <= tick < end_time_step:
                    self._send_buffer.add_key(
                        tick, keys[atom] + (tick & colour_mask))
